Package: mx.client
Type: Package
Title: Stateful Matrix Client Helpers
Version: 0.2.0
Date: 2026-08-04
Authors@R: c(
    person("Troy", "Hernandez", role = c("aut", "cre"),
           email = "troy@cornball.ai",
           comment = c(ORCID = "0009-0005-4248-604X")),
    person("cornball.ai", role = "cph"))
Description: Stateful helpers for building 'Matrix' (<https://matrix.org>)
    chat clients in R. Builds on the low-level 'mx.api' Client-Server API
    bindings, adding local configuration persistence, room resolution,
    sync cursor handling, sync-event extraction, invite acceptance, a
    conservative Markdown-to-HTML converter for formatted messages, and
    'Olm'/'Megolm' end-to-end encryption orchestration over the optional
    'mx.crypto' package.
License: Apache License (>= 2)
URL: https://github.com/cornball-ai/mx.client
BugReports: https://github.com/cornball-ai/mx.client/issues
Depends: R (>= 4.0)
Imports: jsonlite, mx.api (>= 0.3.0), stats, tools, utils
Suggests: mx.crypto (>= 0.2.0), simplermarkdown, tinytest
VignetteBuilder: simplermarkdown
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-08-04 20:22:21 UTC; troy
Author: Troy Hernandez [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-4248-604X>),
  cornball.ai [cph]
Maintainer: Troy Hernandez <troy@cornball.ai>
Repository: CRAN
Date/Publication: 2026-08-04 22:10:08 UTC
Built: R 4.6.0; ; 2026-08-04 23:04:45 UTC; unix

# tinyrox says don't edit this manually, but it can't stop you!

export(mx_accept_invites)
export(mx_client_config_path)
export(mx_client_configure)
export(mx_client_from_config)
export(mx_client_legacy_config_path)
export(mx_client_load)
export(mx_client_relogin)
export(mx_client_save)
export(mx_client_session)
export(mx_crypto_account)
export(mx_crypto_account_save)
export(mx_crypto_claim_otks)
export(mx_crypto_decrypt_event)
export(mx_crypto_device_keys)
export(mx_crypto_encrypt_event)
export(mx_crypto_encrypt_for_devices)
export(mx_crypto_handle_to_device)
export(mx_crypto_inbound_session)
export(mx_crypto_known_devices)
export(mx_crypto_process_sync)
export(mx_crypto_publish_keys)
export(mx_crypto_room_key_payload)
export(mx_crypto_sessions_load)
export(mx_crypto_sessions_new)
export(mx_crypto_sessions_save)
export(mx_crypto_store_dir)
export(mx_extract_invites)
export(mx_extract_reaction_verdict)
export(mx_extract_text_events)
export(mx_markdown_to_html)
export(mx_pill_mentions)
export(mx_resolve_room)
export(mx_room_encrypted)
export(mx_room_lookup_by_name)
export(mx_send_encrypted)
export(mx_send_media)
export(mx_send_table)
export(mx_send_text)
export(mx_set_displayname)
export(mx_sync_update)
export(mx_table_html)
export(mx_with_relogin)

S3method(print,mx_client_config)

# mx.client

Stateful Matrix client helpers for R.

`mx.api` owns the raw Matrix Client-Server HTTP endpoints. `mx.crypto`
owns the Olm and Megolm cryptographic primitives. `mx.client` is the
layer between them: local configuration, room resolution, sync cursor
handling, event extraction, invite acceptance, markdown-formatted
messages, and end-to-end encryption orchestration.

## Install

```r
# From GitHub while the CRAN submission is in flight:
remotes::install_github("cornball-ai/mx.client")

# Optional, for end-to-end encryption (needs a Rust toolchain to build):
install.packages("mx.crypto")
```

## Quick start: plaintext

```r
library(mx.client)

# One-time setup: log in, join the room, persist credentials
# (mode 0600, under tools::R_user_dir()).
mx_client_configure(
    server = "https://matrix.example.org",
    user = "bot",
    password = "secret",
    room = "#general:example.org",
    app = "myapp"
)

# Every later session:
client <- mx_client_load(app = "myapp")
mx_send_text(client, "hello from R")                       # default room
mx_send_text(client, "**bold** and `code`", room = "general",
             markdown = TRUE)                              # named room, HTML

# Poll for new messages and advance the stored sync cursor:
res <- mx_sync_update(client, timeout = 30000L)
msgs <- mx_extract_text_events(res$sync, client$user_id)
```

## Media, and surviving token rotation

Both need `mx.api (>= 0.3.0)`, which mx.client now pins.

```r
# Send a file with the stored client; the room resolves by name and the
# msgtype comes from the MIME type (a .png posts as m.image).
mx_send_media(client, "plot.png", room = "general")

# Recover from an invalidated access token without babysitting:
# catches M_UNKNOWN_TOKEN, re-logs in with the stored password
# (preserving the device id, so an E2EE identity survives), saves the
# refreshed token, and retries once.
mx_with_relogin(client, function(cl) {
    mx_send_text(cl, "still here after a token rotation")
})
```

## End-to-end encryption

Olm/Megolm send/receive orchestration over `mx.crypto`, aimed at bots
and controlled deployments. `mx.crypto` is a Suggests and is only called
from the E2EE entry points, so plaintext Matrix clients install and run
without a Rust toolchain.

The mental model: an *account* holds this device's identity keys, the
*store* persists everything on disk, *sessions* hold the live Olm (per
peer device) and Megolm (per room) ratchet state, and a *room key* is
what lets a device decrypt a room. mx.client moves all four around for
you.

```r
# Once: create the crypto store and publish this device's keys.
store <- mx_crypto_store_dir("myapp")
acct <- mx_crypto_account(store)
mx_crypto_publish_keys(client, acct, store)

# Send into an encrypted room. Device discovery, one-time-key claims,
# and Megolm key sharing happen inside.
sessions <- mx_crypto_sessions_load(store)
room_id <- mx_resolve_room(client, "secret plans")   # or a literal !id
res <- mx_send_encrypted(client, acct, sessions, room_id,
                         list(msgtype = "m.text", body = "secret"),
                         store, member_ids = "@friend:example.org")
sessions <- res$sessions

# Receive: recover room keys from to-device messages and decrypt
# timeline events in one pass.
my_curve <- mx.crypto::mxc_account_identity_keys(acct)$curve25519
sync <- mx_sync_update(client)$sync
out <- mx_crypto_process_sync(acct, sessions, sync,
                              my_curve, self_id = client$user_id)
out$events   # same shape as mx_extract_text_events()
```

Security model, in brief: device keys are trusted on first use (no
cross-signing trust store yet), and there is no key-request or
forwarded-key flow. See `vignette("e2ee", package = "mx.client")` for
the full flow, what happens on the wire, and the current limitations.

## The package family

| Package | Role | Depends on |
|---|---|---|
| [mx.api](https://github.com/cornball-ai/mx.api) | HTTP transport (CRAN) | — |
| [mx.crypto](https://github.com/cornball-ai/mx.crypto) | Olm/Megolm primitives (vodozemac) | — |
| mx.client | stateful client + E2EE orchestration | Imports mx.api, Suggests mx.crypto |

# Room lookup and resolution.

#' Look up a joined room by display name
#'
#' @param client Matrix client config.
#' @param name Character room name.
#' @return Room id, or NULL when no joined room has that name.
#' @examples
#' \dontrun{
#' # Needs a live homeserver session.
#' mx_room_lookup_by_name(mx_client_load("myapp"), "general")
#' }
#' @export
mx_room_lookup_by_name <- function(client, name) {
    if (!is.character(name) || length(name) != 1L || !nzchar(name)) {
        stop("name must be a non-empty character scalar", call. = FALSE)
    }
    s <- mx_client_session(client)
    rooms <- tryCatch(mx.api::mx_rooms(s), error = function(e) character())
    for (rid in rooms) {
        nm <- tryCatch(mx.api::mx_room_name(s, rid), error = function(e) NULL)
        if (!is.null(nm) && identical(nm, name)) {
            return(rid)
        }
    }
    NULL
}

#' Resolve a room id, name, or default room
#'
#' Resolution order:
#' literal room IDs beginning with \code{!}, a supplied \code{room_cache}
#' name-to-id map, joined-room display-name lookup, then the config's
#' \code{room_id} fallback.
#'
#' @param client Matrix client config.
#' @param room Character or NULL.
#' @param room_cache Named list or character vector mapping names to ids.
#' @param fallback Logical. Use \code{client$room_id} when lookup misses.
#' @param details Logical. Return source metadata instead of just the id.
#' @param quiet Logical. Suppress fallback message.
#' @return Character room id, or a list when \code{details = TRUE}.
#' @examples
#' client <- list(room_id = "!default:example.org")
#' # Literal ids and cache hits resolve without a server round-trip:
#' mx_resolve_room(client, "!abc:example.org")
#' mx_resolve_room(client, "general",
#'                 room_cache = list(general = "!gen:example.org"))
#' mx_resolve_room(client) # NULL room falls back to the config default
#' @export
mx_resolve_room <- function(client, room = NULL, room_cache = NULL,
                            fallback = TRUE, details = FALSE, quiet = FALSE) {
    done <- function(room_id, source) {
        if (isTRUE(details)) {
            list(room_id = room_id, source = source)
        } else {
            room_id
        }
    }

    default_room <- client$room_id
    if (is.null(room) || !nzchar(room)) {
        if (is.null(default_room) || !nzchar(default_room)) {
            stop("room is NULL and client has no default room_id",
                 call. = FALSE)
        }
        return(done(default_room, "default"))
    }
    if (!is.character(room) || length(room) != 1L) {
        stop("room must be NULL or a character scalar", call. = FALSE)
    }
    if (startsWith(room, "!")) {
        return(done(room, "literal"))
    }

    cached <- room_cache[[room]]
    if (!is.null(cached) && length(cached) == 1L && nzchar(cached)) {
        return(done(unname(cached), "cache"))
    }

    rid <- mx_room_lookup_by_name(client, room)
    if (!is.null(rid)) {
        return(done(rid, "lookup"))
    }

    if (!isTRUE(fallback)) {
        stop(sprintf("No joined room named %s", sQuote(room)), call. = FALSE)
    }
    if (is.null(default_room) || !nzchar(default_room)) {
        stop(sprintf("No joined room named %s and no default room_id is set",
                     sQuote(room)), call. = FALSE)
    }
    if (!isTRUE(quiet)) {
        message(sprintf("mx.client: no joined room named %s; using default",
                        room))
    }
    done(default_room, "fallback")
}

#' Is a room end-to-end encrypted?
#'
#' Resolves the room (by name, id, or the config default) and reads its
#' \code{m.room.encryption} state. Needs mx.api >= 0.3.0.
#'
#' @param client Matrix client config.
#' @param room Character room id/name or NULL for the default room.
#' @param room_cache Optional room name-to-id cache.
#' @return TRUE when the room advertises an encryption algorithm,
#'   FALSE otherwise.
#' @examples
#' \dontrun{
#' if (mx_room_encrypted(client, "secret plans")) {
#'     # use mx_send_encrypted() instead of mx_send_text()
#' }
#' }
#' @export
mx_room_encrypted <- function(client, room = NULL, room_cache = NULL) {
    rid <- mx_resolve_room(client, room, room_cache = room_cache)
    enc <- mx.api::mx_get_state(mx_client_session(client), rid,
                                "m.room.encryption")
    !is.null(enc$algorithm)
}

# Profile operations: client-level wrappers over mx.api's profile
# endpoints, so callers hand in the client config and get token
# rotation handled instead of building sessions by hand.

#' Set the account's profile display name
#'
#' Client-level wrapper over \code{mx.api::mx_set_displayname()}: builds
#' the session from the client config and retries once through
#' \code{\link{mx_with_relogin}} when the homeserver rejects a rotated
#' token. The display name is account-global and shows in the sender
#' line of every message the account posts.
#'
#' @param client Matrix client config.
#' @param name Character. New display name.
#' @param save Logical. On a relogin, persist the refreshed token to the
#'   client's config path (see \code{\link{mx_client_relogin}}).
#' @return TRUE, invisibly.
#' @examples
#' \dontrun{
#' # A real rename needs a live homeserver session:
#' client <- mx_client_load("myapp")
#' mx_set_displayname(client, "mybot (maintenance)")
#' }
#' @export
mx_set_displayname <- function(client, name, save = TRUE) {
    if (!is.character(name) || length(name) != 1L || is.na(name) ||
            !nzchar(name)) {
        stop("name must be a single non-empty string", call. = FALSE)
    }
    mx_with_relogin(client, function(cl) {
        mx.api::mx_set_displayname(mx_client_session(cl), name)
    }, save = save)
    invisible(TRUE)
}

library(tinytest)
library(mx.client)

# Signatures
expect_equal(names(formals(mx_client_relogin)), c("client", "save"))
expect_equal(names(formals(mx_with_relogin)), c("client", "fn", "save"))
expect_equal(
  names(formals(mx.client::mx_send_media)),
  c("client", "path", "room", "body", "msgtype", "content_type",
    "info", "room_cache", "dry_run")
)

cl <- mx_client_from_config(list(server = "https://x", user = "bot",
                                 token = "t", user_id = "@bot:x",
                                 device_id = "DEV", room_id = "!r:x"))

# mx_with_relogin: success path never relogs
expect_equal(mx_with_relogin(cl, function(c) "ok"), "ok")

# unrelated errors propagate untouched
expect_error(mx_with_relogin(cl, function(c) stop("boom")), "boom")

# token error without a stored password -> clear failure from relogin
token_err <- function(c) {
  stop(structure(class = c("mx_error_M_UNKNOWN_TOKEN", "mx_error",
                           "error", "condition"),
                 list(message = "Matrix error [M_UNKNOWN_TOKEN]: x",
                      call = NULL)))
}
expect_error(mx_with_relogin(cl, token_err), "no stored password")

# mx_send_media dry-run resolves the default room and sends nothing
expect_message(
  out <- mx.client::mx_send_media(cl, "clip.mp4", dry_run = TRUE),
  "dry-run"
)
expect_null(out)

# mx_set_displayname: signature, validation, and relogin plumbing
expect_equal(names(formals(mx.client::mx_set_displayname)),
             c("client", "name", "save"))
expect_error(mx.client::mx_set_displayname(cl, ""), "non-empty")
expect_error(mx.client::mx_set_displayname(cl, c("a", "b")), "single")
expect_error(mx.client::mx_set_displayname(cl, NA_character_), "non-empty")

# Package-level documentation.

#' mx.client: Stateful Matrix Client Helpers
#'
#' Stateful helpers for building 'Matrix' chat clients in R, layered on
#' the low-level \pkg{mx.api} Client-Server API bindings: local
#' configuration persistence, room resolution, sync cursor handling,
#' sync-event extraction, invite acceptance, a conservative
#' Markdown-to-HTML converter, and Olm/Megolm end-to-end encryption
#' orchestration over the optional \pkg{mx.crypto} package.
#'
#' @seealso \code{\link{mx_client_configure}} to create a config,
#'   \code{\link{mx_client_load}} / \code{\link{mx_client_session}} to use
#'   one, \code{\link{mx_send_text}} and \code{\link{mx_sync_update}} for
#'   the core send/receive loop.
#' @name mx.client-package
#' @aliases mx.client
"_PACKAGE"

# Message send, sync cursor updates, and sync event extraction.

#' Send plain text to a Matrix room
#'
#' @param client Matrix client config.
#' @param text Character message body.
#' @param room Character room id/name or NULL for the default room.
#' @param msgtype Character Matrix message type.
#' @param room_cache Optional room name-to-id cache.
#' @param dry_run Logical. Print instead of sending.
#' @param markdown Logical. If TRUE, include Matrix custom HTML derived
#'   from a conservative markdown subset.
#' @param mentions Character vector of Matrix user ids to mention
#'   (e.g. \code{"@jorge:cornball.ai"}). Each id is added to the event's
#'   \code{m.mentions} (so the user is notified) and any textual
#'   \code{@localpart} in the body becomes a \code{matrix.to} pill in the
#'   HTML. Implies an HTML formatted body even when \code{markdown} is
#'   FALSE -- pills only render from HTML.
#' @return Event id, or NULL on dry-run.
#' @examples
#' client <- list(room_id = "!default:example.org")
#' mx_send_text(client, "release is out", dry_run = TRUE)
#' \dontrun{
#' # A real send needs a live homeserver session:
#' client <- mx_client_load("myapp")
#' mx_send_text(client, "release is out", markdown = TRUE,
#'              mentions = "@jorge:example.org")
#' }
#' @export
mx_send_text <- function(client, text, room = NULL, msgtype = "m.text",
                         room_cache = NULL, dry_run = FALSE,
                         markdown = FALSE, mentions = NULL) {
    rid <- mx_resolve_room(client, room, room_cache = room_cache)
    if (isTRUE(dry_run)) {
        message("=== mx_send_text (dry-run) [", room %||% "default",
                "] ===\n", text)
        return(invisible(NULL))
    }
    extra <- NULL
    if (isTRUE(markdown) || length(mentions)) {
        html <- mx_markdown_to_html(text)
        if (length(mentions)) {
            html <- mx_pill_mentions(html, mentions)
        }
        extra <- list(format = "org.matrix.custom.html", formatted_body = html)
    }
    if (length(mentions)) {
        extra <- c(extra,
                   list("m.mentions" = list(user_ids = as.list(mentions))))
    }
    mx.api::mx_send(mx_client_session(client), rid, text, msgtype = msgtype,
                    extra = extra)
}

#' Sync once and update the stored cursor
#'
#' Calls \code{mx.api::mx_sync()} using \code{client$sync_token}, stores
#' the returned \code{next_batch} in a returned client object, and
#' optionally saves it back to disk.
#'
#' @param client Matrix client config.
#' @param timeout Integer long-poll timeout in milliseconds.
#' @param filter Character or NULL. Matrix sync filter.
#' @param save Logical. Persist the updated client config.
#' @param path Character or NULL. Save destination.
#' @param app Character or NULL. Application namespace for default saves.
#' @return List with \code{sync}, \code{client}, and \code{first_run}.
#' @examples
#' \dontrun{
#' # Needs a live homeserver session.
#' client <- mx_client_load("myapp")
#' res <- mx_sync_update(client, timeout = 30000L)
#' events <- mx_extract_text_events(res$sync, client$user_id)
#' }
#' @export
mx_sync_update <- function(client, timeout = 0L, filter = NULL, save = TRUE,
                           path = NULL, app = NULL) {
    first_run <- is.null(client$sync_token)
    sync <- mx.api::mx_sync(mx_client_session(client),
                            since = client$sync_token,
                            timeout = as.integer(timeout), filter = filter)
    updated <- mx_client_from_config(
                                     mx_client_plain_list(client),
                                     path = path %||% attr(client, "path"),
                                     app = app %||% attr(client, "app")
    )
    updated$sync_token <- sync$next_batch
    if (isTRUE(save)) {
        updated <- mx_client_save(updated, app = app, path = path)
    }
    list(sync = sync, client = updated, first_run = first_run)
}

#' Extract text message events from a sync response
#'
#' Walks joined-room timeline events and returns normalized text-message
#' records. Self events are retained and tagged with \code{is_self}.
#'
#' @param sync_resp Parsed \code{/sync} response.
#' @param self_id Current user's Matrix id.
#' @param msgtypes Character vector of message types to include.
#' @return List of normalized event records, each carrying
#'   \code{room_id}, \code{event_id}, \code{sender}, \code{is_self},
#'   \code{body}, \code{msgtype}, \code{ts} (the event's
#'   \code{origin_server_ts}, in milliseconds since the epoch, or NULL
#'   when the server omits it), and \code{mentions}.
#' @examples
#' sync_resp <- list(rooms = list(join = list("!room:example.org" = list(
#'     timeline = list(events = list(list(type = "m.room.message",
#'         event_id = "$1", sender = "@alice:example.org",
#'         content = list(msgtype = "m.text", body = "hello"))))))))
#' mx_extract_text_events(sync_resp, self_id = "@bot:example.org")
#' @export
mx_extract_text_events <- function(sync_resp, self_id, msgtypes = "m.text") {
    joined <- sync_resp$rooms$join
    if (!length(joined)) {
        return(list())
    }

    out <- list()
    for (rid in names(joined)) {
        events <- joined[[rid]]$timeline$events
        if (!length(events)) {
            next
        }
        for (ev in events) {
            if (isTRUE(ev$type == "m.room.message") &&
                isTRUE(ev$content$msgtype %in% msgtypes) &&
                !is.null(ev$content$body)) {
                out[[length(out) + 1L]] <- list(room_id = rid,
                    event_id = ev$event_id, sender = ev$sender,
                    is_self = isTRUE(ev$sender == self_id),
                    body = ev$content$body, msgtype = ev$content$msgtype,
                    ts = ev$origin_server_ts,
                    mentions = ev$content$`m.mentions`$user_ids)
            }
        }
    }
    out
}

#' Extract pending invite room ids from a sync response
#'
#' @param sync_resp Parsed \code{/sync} response.
#' @return Character vector of invited room ids.
#' @examples
#' sync_resp <- list(rooms = list(invite = list("!inv:example.org" = list())))
#' mx_extract_invites(sync_resp)
#' @export
mx_extract_invites <- function(sync_resp) {
    invited <- sync_resp$rooms$invite
    if (!length(invited)) {
        return(character())
    }
    names(invited)
}

#' Accept pending Matrix room invites
#'
#' @param client Matrix client config.
#' @param invites Character vector of room ids.
#' @return Character vector of joined room ids.
#' @examples
#' \dontrun{
#' # Needs a live homeserver session.
#' client <- mx_client_load("myapp")
#' res <- mx_sync_update(client)
#' mx_accept_invites(res$client, mx_extract_invites(res$sync))
#' }
#' @export
mx_accept_invites <- function(client, invites) {
    if (!length(invites)) {
        return(character())
    }
    s <- mx_client_session(client)
    joined <- character()
    for (rid in invites) {
        out <- tryCatch(
                        mx.api::mx_room_join(s, rid),
                        error = function(e) {
            message(sprintf("mx.client: failed to join %s: %s", rid,
                            conditionMessage(e)))
            NULL
        }
        )
        if (!is.null(out)) {
            joined <- c(joined, out)
        }
    }
    joined
}

#' Extract a reaction approval verdict from sync events
#'
#' Scans a room timeline for a reaction on \code{target_event_id} from
#' someone other than \code{self_id}. Returns TRUE for approval keys,
#' FALSE for denial keys, or NULL when no verdict is present.
#'
#' @param sync_resp Parsed \code{/sync} response.
#' @param room_id Character room id.
#' @param self_id Current user's Matrix id.
#' @param target_event_id Event id being reacted to.
#' @param approve_keys Character vector of reaction keys read as approval.
#'   \code{NULL} (default) uses thumbs-up (U+1F44D), check-mark (U+2705),
#'   and \code{"y"}/\code{"yes"}/\code{"ok"}.
#' @param deny_keys Character vector of reaction keys read as denial.
#'   \code{NULL} (default) uses thumbs-down (U+1F44E), cross-mark
#'   (U+274C), and \code{"n"}/\code{"no"}/\code{"nope"}.
#' @return TRUE, FALSE, or NULL.
#' @examples
#' sync_resp <- list(rooms = list(join = list("!room:example.org" = list(
#'     timeline = list(events = list(list(type = "m.reaction",
#'         sender = "@alice:example.org",
#'         content = list("m.relates_to" = list(rel_type = "m.annotation",
#'             event_id = "$msg", key = "yes")))))))))
#' mx_extract_reaction_verdict(sync_resp, "!room:example.org",
#'                             self_id = "@bot:example.org",
#'                             target_event_id = "$msg")
#' @export
mx_extract_reaction_verdict <- function(sync_resp, room_id, self_id,
                                        target_event_id, approve_keys = NULL,
                                        deny_keys = NULL) {
    # Emoji defaults are built here, not in the signature, so they don't
    # land as raw astral-plane glyphs in the .Rd \usage block -- LaTeX
    # can't typeset them and the PDF manual fails R CMD check --as-cran.
    if (is.null(approve_keys)) {
        approve_keys <- c(intToUtf8(0x1F44D), intToUtf8(0x2705), "y", "yes",
                          "ok")
    }
    if (is.null(deny_keys)) {
        deny_keys <- c(intToUtf8(0x1F44E), intToUtf8(0x274C), "n", "no", "nope")
    }
    room <- sync_resp$rooms$join[[room_id]]
    if (is.null(room)) {
        return(NULL)
    }
    events <- room$timeline$events
    if (!length(events)) {
        return(NULL)
    }

    for (ev in events) {
        if (!isTRUE(ev$type == "m.reaction")) {
            next
        }
        if (isTRUE(ev$sender == self_id)) {
            next
        }
        rel <- ev$content$`m.relates_to`
        if (!is.list(rel) || !identical(rel$event_id, target_event_id)) {
            next
        }
        key <- rel$key
        if (!is.character(key) || !length(key)) {
            next
        }
        if (key %in% approve_keys) {
            return(TRUE)
        }
        if (key %in% deny_keys) {
            return(FALSE)
        }
    }
    NULL
}

#' Send a media file to a Matrix room
#'
#' Client-layer wrapper over \code{mx.api::mx_send_media()}: resolves the
#' room by name (or falls back to the config's default room), builds the
#' session from the client config, and uploads + posts in one call. The
#' msgtype is derived from the file's MIME type unless given.
#'
#' If you attach mx.api and mx.client together, namespace-qualify -- the
#' two packages export an \code{mx_send_media} each (session-first there,
#' client-first here).
#'
#' @param client Matrix client config.
#' @param path Character. Path to the file to upload.
#' @param room Character room id/name or NULL for the default room.
#' @param body Character. Message body / filename shown by clients.
#' @param msgtype Character or NULL. NULL derives it from the MIME type.
#' @param content_type Character or NULL. MIME type override for files
#'   whose extension guesses wrong (tempfiles, odd extensions); NULL
#'   guesses from the extension.
#' @param info List. Extra fields merged into the media \code{info}.
#' @param room_cache Optional room name-to-id cache.
#' @param dry_run Logical. Print instead of uploading/sending.
#' @return Event id, or NULL on dry-run.
#' @examples
#' client <- list(room_id = "!default:example.org")
#' png <- file.path(tempdir(), "plot.png")
#' file.create(png)
#' mx_send_media(client, png, dry_run = TRUE)
#' unlink(png)
#' @export
mx_send_media <- function(client, path, room = NULL, body = basename(path),
                          msgtype = NULL, content_type = NULL, info = list(),
                          room_cache = NULL, dry_run = FALSE) {
    rid <- mx_resolve_room(client, room, room_cache = room_cache)
    if (isTRUE(dry_run)) {
        message("=== mx_send_media (dry-run) [", room %||% "default",
                "] ===\n", path, " (",
                content_type %||% mx.api::mx_guess_mime(path), ")")
        return(invisible(NULL))
    }
    mx.api::mx_send_media(mx_client_session(client), rid, path,
                          body = body, msgtype = msgtype,
                          content_type = content_type, info = info)
}

# --- Config paths and persistence ---

tmp <- tempfile("mx-client-test-")
dir.create(tmp, recursive = TRUE)
cfg_path <- file.path(tmp, "matrix.json")
legacy_path <- file.path(tmp, "legacy.json")
old_env <- Sys.getenv("MX_CLIENT_TEST_MATRIX_CONFIG", NA_character_)
Sys.setenv(MX_CLIENT_TEST_MATRIX_CONFIG = cfg_path)

cfg <- mx.client::mx_client_from_config(
    list(server = "https://matrix.example", token = "tok",
         user_id = "@bot:example", device_id = "DEV", room_id = "!room:ex"),
    app = "mx.client.test"
)
saved <- mx.client::mx_client_save(cfg, app = "mx.client.test")
expect_equal(attr(saved, "path"), cfg_path)
expect_true(file.exists(cfg_path))
loaded <- mx.client::mx_client_load(app = "mx.client.test")
expect_equal(loaded$user_id, "@bot:example")
expect_inherits(loaded, "mx_client_config")

Sys.unsetenv("MX_CLIENT_TEST_MATRIX_CONFIG")
writeLines(jsonlite::toJSON(list(server = "s", token = "t",
                                 user_id = "@u:s", device_id = "D",
                                 room_id = "!r:s"),
                            auto_unbox = TRUE), legacy_path)
loaded_legacy <- mx.client::mx_client_load(
    app = "mx.client.test",
    legacy_path = legacy_path
)
expect_equal(loaded_legacy$room_id, "!r:s")

# --- Session validation ---

sess <- mx.client::mx_client_session(loaded)
expect_inherits(sess, "mx_session")
expect_error(mx.client::mx_client_session(list(server = "s")),
             pattern = "missing fields")

# --- Room resolution without network ---

expect_equal(mx.client::mx_resolve_room(loaded, NULL), "!room:ex")
expect_equal(mx.client::mx_resolve_room(loaded, "!literal:ex"), "!literal:ex")
resolved <- mx.client::mx_resolve_room(
    loaded, "Ops", room_cache = list(Ops = "!ops:ex"), details = TRUE
)
expect_equal(resolved$room_id, "!ops:ex")
expect_equal(resolved$source, "cache")

# --- Sync response extraction ---

sync <- list(rooms = list(join = list(
    "!r:ex" = list(timeline = list(events = list(
        list(type = "m.room.message", sender = "@alice:ex",
             event_id = "$1",
             content = list(msgtype = "m.text", body = "hello",
                            `m.mentions` = list(user_ids = list("@bot:ex")))),
        list(type = "m.room.message", sender = "@bot:ex",
             event_id = "$2",
             content = list(msgtype = "m.text", body = "self")),
        list(type = "m.room.message", sender = "@alice:ex",
             event_id = "$3",
             content = list(msgtype = "m.image", body = "skip"))
    )))
)))
events <- mx.client::mx_extract_text_events(sync, "@bot:ex")
expect_equal(length(events), 2L)
expect_false(events[[1]]$is_self)
expect_true(events[[2]]$is_self)
expect_equal(events[[1]]$mentions[[1]], "@bot:ex")

# origin_server_ts survives extraction. Consumers that stamp messages
# with the poll's wall clock instead reorder history on every restart,
# so a record without it is not merely lossy.
ts_sync <- list(rooms = list(join = list("!r:ex" = list(timeline = list(
    events = list(
        list(type = "m.room.message", sender = "@alice:ex", event_id = "$1",
             origin_server_ts = 1700000000000,
             content = list(msgtype = "m.text", body = "stamped")),
        list(type = "m.room.message", sender = "@alice:ex", event_id = "$2",
             content = list(msgtype = "m.text", body = "unstamped"))
    )
)))))
ts_events <- mx.client::mx_extract_text_events(ts_sync, "@bot:ex")
expect_equal(ts_events[[1]]$ts, 1700000000000)
# A server that omits it yields NULL rather than a fabricated time
expect_null(ts_events[[2]]$ts)

invites <- mx.client::mx_extract_invites(list(rooms = list(invite = list(
    "!a:ex" = list(), "!b:ex" = list()
))))
expect_equal(invites, c("!a:ex", "!b:ex"))
expect_equal(mx.client::mx_extract_invites(list(rooms = list(join = list()))),
             character())

# --- Reaction verdict extraction ---

reaction_sync <- list(rooms = list(join = list(
    "!r:ex" = list(timeline = list(events = list(
        list(type = "m.reaction", sender = "@bot:ex",
             content = list(`m.relates_to` = list(event_id = "$target",
                                                  key = "\U0001F44D"))),
        list(type = "m.reaction", sender = "@alice:ex",
             content = list(`m.relates_to` = list(event_id = "$target",
                                                  key = "\U0001F44D")))
    )))
)))
expect_true(mx.client::mx_extract_reaction_verdict(
    reaction_sync, "!r:ex", "@bot:ex", "$target"
))
expect_null(mx.client::mx_extract_reaction_verdict(
    reaction_sync, "!missing:ex", "@bot:ex", "$target"
))

if (is.na(old_env)) {
    Sys.unsetenv("MX_CLIENT_TEST_MATRIX_CONFIG")
} else {
    Sys.setenv(MX_CLIENT_TEST_MATRIX_CONFIG = old_env)
}
unlink(tmp, recursive = TRUE)

# mx_room_encrypted: signature (network behavior is live-validated)
expect_true(is.function(mx.client::mx_room_encrypted))
expect_equal(names(formals(mx.client::mx_room_encrypted)),
             c("client", "room", "room_cache"))

# --- print.mx_client_config masks credentials ---

secret_cfg <- mx.client::mx_client_from_config(
    list(server = "https://matrix.example", token = "syt_supersecret",
         password = "hunter2", user_id = "@bot:example", device_id = "DEV",
         room_id = "!room:ex", sync_token = "s99_cursor",
         tools_filter = list()),
    path = "/tmp/matrix.json"
)
out <- paste(capture.output(print(secret_cfg)), collapse = "\n")
# credentials never appear, in any form
expect_false(grepl("syt_supersecret", out, fixed = TRUE))
expect_false(grepl("hunter2", out, fixed = TRUE))
expect_true(grepl("token:\\s+<hidden>", out))
expect_true(grepl("password:\\s+<hidden>", out))
# non-secret fields stay visible, including the sync cursor
expect_true(grepl("@bot:example", out, fixed = TRUE))
expect_true(grepl("s99_cursor", out, fixed = TRUE))
expect_true(grepl("/tmp/matrix.json", out, fixed = TRUE))
# an empty credential reads as unset rather than hidden
empty_cfg <- mx.client::mx_client_from_config(
    list(server = "https://matrix.example", token = "", user_id = "@b:e")
)
out_empty <- paste(capture.output(print(empty_cfg)), collapse = "\n")
expect_true(grepl("token:\\s+<unset>", out_empty))
# printing returns the object unchanged, and invisibly
capture.output(vis <- withVisible(print(secret_cfg)))
expect_false(vis$visible)
expect_identical(vis$value, secret_cfg)

# Local Matrix client configuration.

`%||%` <- function(x, y) {
    if (is.null(x) || !length(x)) {
        y
    } else {
        x
    }
}

mx_client_envvar <- function(app) {
    app <- gsub("[^A-Za-z0-9]+", "_", app)
    paste0(toupper(app), "_MATRIX_CONFIG")
}

#' Path to a Matrix client config file
#'
#' Resolves the config path for an application that uses mx.client. The
#' default environment variable is derived from \code{app}; for example,
#' \code{app = "corteza"} honors \code{CORTEZA_MATRIX_CONFIG}.
#'
#' @param app Character. Application namespace for \code{tools::R_user_dir()}.
#' @param env_var Character or NULL. Override environment variable name.
#' @return Character path.
#' @examples
#' mx_client_config_path("myapp")
#' @export
mx_client_config_path <- function(app = "mx.client", env_var = NULL) {
    env_var <- env_var %||% mx_client_envvar(app)
    env <- Sys.getenv(env_var, "")
    if (nzchar(env)) {
        return(path.expand(env))
    }
    file.path(tools::R_user_dir(app, "config"), "matrix.json")
}

#' Legacy Matrix config path for an application
#'
#' Currently only \code{app = "corteza"} has a historical path:
#' \code{~/.corteza/matrix.json}.
#'
#' @param app Character. Application namespace.
#' @return Character path or NULL.
#' @examples
#' mx_client_legacy_config_path("corteza")
#' @export
mx_client_legacy_config_path <- function(app = "mx.client") {
    if (identical(app, "corteza")) {
        return(path.expand("~/.corteza/matrix.json"))
    }
    NULL
}

#' Wrap a list as an mx.client config
#'
#' @param cfg Named list.
#' @param path Character or NULL. Source/sink path for saves.
#' @param app Character or NULL. Application namespace.
#' @return An object of class \code{"mx_client_config"}.
#' @examples
#' cfg <- mx_client_from_config(list(server = "https://matrix.example.org",
#'                                   token = "syt_example",
#'                                   user_id = "@bot:example.org",
#'                                   device_id = "DEVICEID"))
#' class(cfg)
#' @export
mx_client_from_config <- function(cfg, path = NULL, app = NULL) {
    if (!is.list(cfg)) {
        stop("cfg must be a list", call. = FALSE)
    }
    structure(cfg, class = unique(c("mx_client_config", class(cfg))),
              path = path, app = app)
}

mx_client_plain_list <- function(client) {
    out <- unclass(client)
    attr(out, "path") <- NULL
    attr(out, "app") <- NULL
    out
}

# Config fields holding credentials. sync_token is deliberately absent:
# it is a cursor, not a secret, and seeing it is useful when debugging.
mx_client_secret_fields <- c("token", "access_token", "refresh_token",
                             "password")

mx_client_field_text <- function(name, value) {
    set <- length(value) > 0L &&
    !all(is.na(value)) &&
    (!is.character(value) || any(nzchar(value)))
    if (name %in% mx_client_secret_fields) {
        return(if (set) {
                "<hidden>"
            } else {
                "<unset>"
            })
    }
    if (!set) {
        return("<unset>")
    }
    if (is.list(value) || length(value) > 1L) {
        return(sprintf("<%s[%d]>", class(value)[1L], length(value)))
    }
    as.character(value)
}

#' Print a Matrix client config
#'
#' Prints the config field by field with credentials masked, so that an
#' interactive session, a screenshot, or a pasted bug report does not
#' leak the access token or password. Secret fields show whether they
#' are set (\code{<hidden>}) or empty (\code{<unset>}) without showing
#' the value. Use \code{unclass(x)} or \code{str(unclass(x))} when the
#' raw credentials are genuinely needed.
#'
#' @param x An \code{mx_client_config}, as returned by
#'   \code{mx_client_load()} or \code{mx_client_from_config()}.
#' @param ... Ignored.
#' @return \code{x}, invisibly.
#' @examples
#' cfg <- mx_client_from_config(list(server = "https://matrix.example.org",
#'                                   token = "syt_secret_value",
#'                                   user_id = "@bot:example.org",
#'                                   device_id = "DEVICEID"))
#' cfg
#' @export
print.mx_client_config <- function(x, ...) {
    cat("<mx_client_config>\n")
    fields <- names(x)
    labels <- c(paste0(fields, ":"), if (!is.null(attr(x, "path"))) "path:")
    if (!length(labels)) {
        cat("  (no fields)\n")
        return(invisible(x))
    }
    width <- max(nchar(labels))
    for (f in fields) {
        cat(sprintf("  %-*s %s\n", width, paste0(f, ":"),
                    mx_client_field_text(f, x[[f]])))
    }
    if (!is.null(attr(x, "path"))) {
        cat(sprintf("  %-*s %s\n", width, "path:", attr(x, "path")))
    }
    invisible(x)
}

#' Load a Matrix client config
#'
#' Reads a JSON config. If \code{path} or the derived environment variable
#' is explicit, that path is authoritative. Otherwise \code{legacy_path}
#' is used as a compatibility fallback when present.
#'
#' @param app Character. Application namespace.
#' @param path Character or NULL. Explicit config path.
#' @param legacy_path Character or NULL. Backward-compatible fallback path.
#' @param env_var Character or NULL. Override environment variable name.
#' @return An \code{"mx_client_config"} object.
#' @examples
#' path <- file.path(tempdir(), "matrix.json")
#' cfg <- mx_client_from_config(list(server = "https://matrix.example.org",
#'                                   token = "syt_example",
#'                                   user_id = "@bot:example.org",
#'                                   device_id = "DEVICEID"))
#' mx_client_save(cfg, path = path)
#' mx_client_load(path = path)$user_id
#' unlink(path)
#' @export
mx_client_load <- function(app = "mx.client", path = NULL,
                           legacy_path = mx_client_legacy_config_path(app),
                           env_var = NULL) {
    env_var <- env_var %||% mx_client_envvar(app)
    env <- Sys.getenv(env_var, "")
    explicit <- !is.null(path) || nzchar(env)
    if (is.null(path)) {
        path <- mx_client_config_path(app, env_var = env_var)
    } else {
        path <- path.expand(path)
    }

    legacy_path <- if (is.null(legacy_path)) {
        NULL
    } else {
        path.expand(legacy_path)
    }
    src <- if (file.exists(path)) {
        path
    } else if (!explicit && !is.null(legacy_path) && file.exists(legacy_path)) {
        legacy_path
    } else {
        stop("Matrix not configured. Create a config with mx_client_configure().",
             call. = FALSE)
    }
    cfg <- jsonlite::fromJSON(src, simplifyVector = TRUE)
    mx_client_from_config(cfg, path = src, app = app)
}

#' Save a Matrix client config
#'
#' Writes JSON with mode 0600.
#'
#' @param client Named list or \code{"mx_client_config"}.
#' @param app Character or NULL. Application namespace.
#' @param path Character or NULL. Destination path.
#' @return The saved config, invisibly.
#' @examples
#' path <- file.path(tempdir(), "matrix.json")
#' mx_client_save(list(server = "https://matrix.example.org",
#'                     token = "syt_example",
#'                     user_id = "@bot:example.org",
#'                     device_id = "DEVICEID"),
#'                path = path)
#' unlink(path)
#' @export
mx_client_save <- function(client, app = NULL, path = NULL) {
    app <- app %||% attr(client, "app") %||% "mx.client"
    path <- path %||% attr(client, "path") %||% mx_client_config_path(app)
    path <- path.expand(path)
    dir.create(dirname(path), showWarnings = FALSE, recursive = TRUE)
    cfg <- mx_client_plain_list(client)
    writeLines(jsonlite::toJSON(cfg, auto_unbox = TRUE, pretty = TRUE), path)
    Sys.chmod(path, mode = "0600")
    invisible(mx_client_from_config(cfg, path = path, app = app))
}

# E2EE transport: the network glue between the crypto state (crypto.R,
# e2ee.R) and mx.api's keys / to-device / send endpoints. These functions
# make real HTTP calls, so they are exercised by a live bring-up rather
# than the in-process loopback tests.

# Sign one published OTK for upload: {key} -> {key, signatures}.
mx_crypto_sign_otk <- function(account, key, user_id, device_id) {
    obj <- list(key = key)
    sig <- mx.crypto::mxc_account_sign(account, mx.api::mx_canonical_json(obj))
    obj$signatures <- stats::setNames(
                                      list(stats::setNames(list(sig), paste0("ed25519:", device_id))),
                                      user_id)
    obj
}

#' Publish this device's identity and one-time keys
#'
#' Builds and signs the device keys and a batch of one-time keys, uploads
#' them with \code{mx.api::mx_keys_upload()}, marks them published, and
#' persists the account. Call once after login and again to replenish
#' one-time keys.
#'
#' @param client Matrix client config (needs \code{user_id},
#'   \code{device_id}).
#' @param account An mx.crypto account handle.
#' @param store_dir Character. Crypto store directory.
#' @param n_otks Integer. Number of one-time keys to publish.
#' @return The \code{/keys/upload} response, invisibly.
#' @examples
#' \dontrun{
#' acct <- mx_crypto_account(mx_crypto_store_dir("corteza"))
#' mx_crypto_publish_keys(mx_client_load(app = "corteza"), acct,
#'                        mx_crypto_store_dir("corteza"))
#' }
#' @export
mx_crypto_publish_keys <- function(client, account, store_dir, n_otks = 50L) {
    mx_require_crypto()
    s <- mx_client_session(client)
    dk <- mx_crypto_device_keys(account, client$user_id, client$device_id)
    mx.crypto::mxc_account_generate_one_time_keys(account, as.integer(n_otks))
    otks <- mx.crypto::mxc_account_one_time_keys(account)
    signed <- stats::setNames(
                              lapply(names(otks), function(id) {
        mx_crypto_sign_otk(account, otks[[id]], client$user_id,
                           client$device_id)
    }),
                              paste0("signed_curve25519:", names(otks)))
    resp <- mx.api::mx_keys_upload(s, device_keys = dk, one_time_keys = signed)
    mx.crypto::mxc_account_mark_published(account)
    mx_crypto_account_save(account, store_dir)
    invisible(resp)
}

#' List the devices (and identity keys) of some users
#'
#' Queries \code{/keys/query} and flattens the result to a list of devices.
#'
#' Devices are returned only if their Ed25519 self-signature verifies
#' against the identity the homeserver claims for them. A device that
#' fails is dropped with a warning naming it, and the remaining devices
#' are still returned: one bad device must not make a room unusable.
#'
#' @param client Matrix client config.
#' @param user_ids Character vector of Matrix user ids.
#' @return List of verified devices, each \code{list(user_id, device_id,
#'   curve25519, ed25519)}.
#' @examples
#' \dontrun{
#' mx_crypto_known_devices(client, "@bob:example.org")
#' }
#' @export
mx_crypto_known_devices <- function(client, user_ids) {
    mx_require_crypto()
    s <- mx_client_session(client)
    query <- stats::setNames(rep(list(list()), length(user_ids)), user_ids)
    resp <- mx.api::mx_keys_query(s, device_keys = query)
    mx_crypto_verify_device_map(resp$device_keys)
}

# Verify every device in a parsed /keys/query device_keys map.
#
# Split out from the HTTP call so the verification logic is testable
# against fixture responses without a homeserver. The homeserver is not
# trusted here: without this check it can substitute its own Curve25519
# key for any device and read everything sent to that user.
#
# mxc_verify_device_keys() raises on failure rather than returning FALSE,
# and returns the keys it actually checked -- use those, not the raw map.
mx_crypto_verify_device_map <- function(device_keys_map) {
    out <- list()
    for (uid in names(device_keys_map %||% list())) {
        devs <- device_keys_map[[uid]]
        for (dev in names(devs %||% list())) {
            keys <- tryCatch(
                             mx.crypto::mxc_verify_device_keys(devs[[dev]], uid, dev),
                             error = function(e) {
                warning("mx.client: skipping unverified device ", uid, "/",
                        dev, ": ", conditionMessage(e), call. = FALSE)
                NULL
            })
            if (is.null(keys)) {
                next
            }
            out[[length(out) + 1L]] <- list(user_id = uid, device_id = dev,
                curve25519 = keys$curve25519, ed25519 = keys$ed25519)
        }
    }
    out
}

#' Claim a one-time key for each device
#'
#' Calls \code{/keys/claim} and attaches the claimed key to each device as
#' \code{$otk}, ready for \code{mx_crypto_encrypt_for_devices()}.
#'
#' Each claimed key's signature is checked against the device's Ed25519
#' key, which must itself have come from a verified \code{device_keys}
#' (that is, from \code{mx_crypto_known_devices()}). A key that fails is
#' dropped with a warning and its device comes back with no \code{otk}.
#'
#' @param client Matrix client config.
#' @param devices List of devices from \code{mx_crypto_known_devices()}.
#' @return The devices with an \code{otk} field added where one was
#'   claimed and verified.
#' @examples
#' \dontrun{
#' devs <- mx_crypto_claim_otks(client, mx_crypto_known_devices(client, uid))
#' }
#' @export
mx_crypto_claim_otks <- function(client, devices) {
    mx_require_crypto()
    if (!length(devices)) {
        return(devices)
    }
    s <- mx_client_session(client)
    req <- list()
    for (d in devices) {
        req[[d$user_id]] <- c(req[[d$user_id]] %||% list(),
                              stats::setNames(list("signed_curve25519"), d$device_id))
    }
    resp <- mx.api::mx_keys_claim(s, one_time_keys = req)
    mx_crypto_verify_claimed_otks(devices, resp$one_time_keys %||% list())
}

# Attach verified one-time keys from a parsed /keys/claim response.
#
# Split from the HTTP call to keep it testable. The previous code took
# slot[[1]]$key and dropped the signatures beside it, so a homeserver
# could hand back a one-time key it had minted itself and we would open
# an Olm session to a device it controls.
#
# mxc_verify_one_time_key() deliberately will not look the signing key up
# for us, since doing so would re-trust the same response it is checking.
# That is why d$ed25519 has to come from an already-verified device.
mx_crypto_verify_claimed_otks <- function(devices, claimed) {
    lapply(devices, function(d) {
        slot <- claimed[[d$user_id]][[d$device_id]]
        if (!length(slot) || is.null(names(slot))) {
            return(d)
        }
        if (is.null(d$ed25519) || !nzchar(d$ed25519)) {
            warning("mx.client: no verified ed25519 key for ", d$user_id,
                    "/", d$device_id, "; cannot check its one-time key",
                    call. = FALSE)
            return(d)
        }
        # slot is "signed_curve25519:<id>" -> {key, signatures}; the outer
        # map key is part of what gets verified, so keep it.
        algo_kid <- names(slot)[[1]]
        otk <- tryCatch(
                        mx.crypto::mxc_verify_one_time_key(algo_kid, slot[[1]], d$ed25519,
                d$user_id, d$device_id),
                        error = function(e) {
            warning("mx.client: rejecting one-time key for ", d$user_id,
                    "/", d$device_id, ": ", conditionMessage(e),
                    call. = FALSE)
            NULL
        })
        if (!is.null(otk)) {
            d$otk <- otk
        }
        d
    })
}

#' Send an end-to-end encrypted message to a room
#'
#' Discovers the room members' devices (unless \code{recipients} is given),
#' claims one-time keys for any without an Olm session, shares the room key
#' over to-device, encrypts the content with Megolm, sends the
#' \code{m.room.encrypted} event, and persists the updated sessions.
#'
#' @param client Matrix client config.
#' @param account An mx.crypto account handle.
#' @param sessions A session set (see \code{mx_crypto_sessions_new()}).
#' @param room_id Character room id.
#' @param content Named list. Plaintext event content.
#' @param store_dir Character. Crypto store directory.
#' @param recipients List of recipient devices, or NULL to discover them
#'   from \code{member_ids}.
#' @param member_ids Character vector of room member user ids (used when
#'   \code{recipients} is NULL).
#' @return List with \code{event_id} and the updated \code{sessions}.
#' @examples
#' \dontrun{
#' res <- mx_send_encrypted(client, acct, sessions, "!r:ex",
#'   list(msgtype = "m.text", body = "secret"), store,
#'   member_ids = "@bob:example.org")
#' }
#' @export
mx_send_encrypted <- function(client, account, sessions, room_id, content,
                              store_dir, recipients = NULL, member_ids = NULL) {
    mx_require_crypto()
    s <- mx_client_session(client)
    sender_curve <- mx.crypto::mxc_account_identity_keys(account)$curve25519

    if (is.null(recipients)) {
        devs <- mx_crypto_known_devices(client, member_ids)
        # skip our own device; claim OTKs only where no Olm session exists
        devs <- Filter(function(d) {
            !identical(d$device_id, client$device_id) &&
            !is.null(d$curve25519) && nzchar(d$curve25519)
        }, devs)
        need <- Filter(function(d) is.null(sessions$olm[[d$curve25519]]), devs)
        have <- Filter(function(d) !is.null(sessions$olm[[d$curve25519]]), devs)
        claimed <- mx_crypto_claim_otks(client, need)
        # A device whose one-time key failed verification has no usable otk.
        # Drop it rather than letting encrypt_for_devices abort the send:
        # the remaining devices should still get the message.
        usable <- Filter(function(d) !is.null(d$otk), claimed)
        if (length(usable) < length(claimed)) {
            warning("mx.client: ", length(claimed) - length(usable), " of ",
                    length(claimed), " new devices in ", room_id,
                    " had no usable one-time key and were skipped",
                    call. = FALSE)
        }
        recipients <- c(usable, have)
    }

    out <- mx_crypto_encrypt_for_devices(account, sessions, room_id,
        content, sender_curve, client$device_id, recipients = recipients,
        sender_user_id = client$user_id)
    for (p in out$to_device) {
        messages <- stats::setNames(
                                    list(stats::setNames(list(p$content), p$device_id)), p$user_id)
        mx.api::mx_send_to_device(s, "m.room.encrypted", messages)
    }
    event_id <- mx.api::mx_send_event(s, room_id, "m.room.encrypted", out$event)
    mx_crypto_sessions_save(out$sessions, store_dir)
    list(event_id = event_id, sessions = out$sessions)
}

# Homeserver key verification. These exercise the pure halves of the
# /keys/query and /keys/claim handling against fixture responses, so the
# security-relevant logic is covered without a live homeserver -- which is
# exactly the gap that let the unverified paths ship in the first place.

library(tinytest)

if (!requireNamespace("mx.crypto", quietly = TRUE)) {
    exit_file("mx.crypto not available (needs a Rust toolchain)")
}
library(mx.client)

UID <- "@alice:example.org"
DEV <- "ALICEDEV"

alice <- mx.crypto::mxc_account_new()
mallory <- mx.crypto::mxc_account_new()
alice_idk <- mx.crypto::mxc_account_identity_keys(alice)
mallory_idk <- mx.crypto::mxc_account_identity_keys(mallory)

dk <- mx_crypto_device_keys(alice, UID, DEV)
good_map <- stats::setNames(list(stats::setNames(list(dk), DEV)), UID)

# ---- a well-formed device verifies and yields its real keys ----
devs <- mx.client:::mx_crypto_verify_device_map(good_map)
expect_equal(length(devs), 1L)
expect_equal(devs[[1]]$user_id, UID)
expect_equal(devs[[1]]$device_id, DEV)
expect_equal(devs[[1]]$curve25519, alice_idk$curve25519)
expect_equal(devs[[1]]$ed25519, alice_idk$ed25519)

# ---- a substituted curve25519 key is rejected ----
# This is the attack the whole exercise exists to stop: the homeserver
# swaps in a key it holds the private half of and reads the room.
swapped <- good_map
swapped[[UID]][[DEV]]$keys[[paste0("curve25519:", DEV)]] <-
    mallory_idk$curve25519
expect_warning(res <- mx.client:::mx_crypto_verify_device_map(swapped))
expect_equal(length(res), 0L)

# ---- a stripped signatures block is rejected ----
unsigned <- good_map
unsigned[[UID]][[DEV]]$signatures <- NULL
expect_warning(res <- mx.client:::mx_crypto_verify_device_map(unsigned))
expect_equal(length(res), 0L)

# ---- a device reattributed to another user is rejected ----
reattributed <- stats::setNames(list(stats::setNames(list(dk), DEV)),
                                "@mallory:example.org")
expect_warning(res <- mx.client:::mx_crypto_verify_device_map(reattributed))
expect_equal(length(res), 0L)

# ---- a device reattributed to another device id is rejected ----
renamed <- stats::setNames(list(stats::setNames(list(dk), "OTHERDEV")), UID)
expect_warning(res <- mx.client:::mx_crypto_verify_device_map(renamed))
expect_equal(length(res), 0L)

# ---- one bad device does not take the good ones with it ----
bob <- mx.crypto::mxc_account_new()
bob_dk <- mx_crypto_device_keys(bob, "@bob:example.org", "BOBDEV")
mixed <- c(good_map,
           stats::setNames(list(stats::setNames(list(bob_dk), "BOBDEV")),
                           "@bob:example.org"))
mixed[[UID]][[DEV]]$signatures <- NULL          # alice's device is broken
expect_warning(res <- mx.client:::mx_crypto_verify_device_map(mixed))
expect_equal(length(res), 1L)                   # bob still usable
expect_equal(res[[1]]$user_id, "@bob:example.org")

# ---- empty / absent map is not an error ----
expect_equal(length(mx.client:::mx_crypto_verify_device_map(NULL)), 0L)
expect_equal(length(mx.client:::mx_crypto_verify_device_map(list())), 0L)

# ---- one-time keys: a correctly signed key verifies ----
mx.crypto::mxc_account_generate_one_time_keys(alice, 1L)
otks <- mx.crypto::mxc_account_one_time_keys(alice)
kid <- names(otks)[[1]]
signed <- mx.client:::mx_crypto_sign_otk(alice, otks[[kid]], UID, DEV)
claimed <- stats::setNames(
    list(stats::setNames(
        list(stats::setNames(list(signed), paste0("signed_curve25519:", kid))),
        DEV)),
    UID)

device <- list(user_id = UID, device_id = DEV,
               curve25519 = alice_idk$curve25519, ed25519 = alice_idk$ed25519)
res <- mx.client:::mx_crypto_verify_claimed_otks(list(device), claimed)
expect_equal(res[[1]]$otk, otks[[kid]])

# ---- a one-time key signed by the wrong device is rejected ----
forged <- mx.client:::mx_crypto_sign_otk(mallory, otks[[kid]], UID, DEV)
forged_claimed <- stats::setNames(
    list(stats::setNames(
        list(stats::setNames(list(forged), paste0("signed_curve25519:", kid))),
        DEV)),
    UID)
expect_warning(
    res <- mx.client:::mx_crypto_verify_claimed_otks(list(device),
                                                     forged_claimed))
expect_null(res[[1]]$otk)

# ---- a device with no verified ed25519 cannot have its OTK checked ----
no_ed <- list(user_id = UID, device_id = DEV,
              curve25519 = alice_idk$curve25519, ed25519 = NULL)
expect_warning(
    res <- mx.client:::mx_crypto_verify_claimed_otks(list(no_ed), claimed))
expect_null(res[[1]]$otk)

# ---- a device the server returned nothing for is passed through ----
res <- mx.client:::mx_crypto_verify_claimed_otks(list(device), list())
expect_equal(length(res), 1L)
expect_null(res[[1]]$otk)

# Stateful E2EE integration: session management on top of the crypto.R
# primitives. These functions turn "send this to an encrypted room" and
# "decrypt this sync" into concrete payloads and plaintext, holding the
# Olm and Megolm sessions in between. Transport (claiming one-time keys,
# sending to-device, sending the event) stays with the caller / mx.api;
# this layer is pure crypto state, so it is testable without a server.
#
# A session set is a list with four named maps:
#   olm        peer Curve25519 -> outbound Olm session (we encrypt to them)
#   olm_in     peer Curve25519 -> inbound Olm session  (they encrypt to us)
#   megolm_out room id          -> list(session, shared = peer curves)
#   megolm_in  "room|session_id" -> list(session, sender, sender_ed25519,
#                                        sender_bound)
#
# megolm_in carries the sender identity the Olm payload claimed when the
# key was shared, plus whether that claim was tied to a device whose
# device_keys verified. The claim alone is not evidence -- anyone can open
# an Olm session to us -- so only sender_bound justifies reporting a
# decrypted event as sender-verified. Stores written before this existed
# hold a bare pickle string there and are migrated on load.

#' Create an empty E2EE session set
#'
#' @return A session set: named lists \code{olm}, \code{olm_in},
#'   \code{megolm_out}, \code{megolm_in}.
#' @examples
#' s <- mx_crypto_sessions_new()
#' names(s)
#' @export
mx_crypto_sessions_new <- function() {
    list(olm = list(), olm_in = list(), megolm_out = list(), megolm_in = list())
}

#' Persist a session set to the crypto store
#'
#' Pickles every live session (encrypted at rest with the store key) into
#' \code{sessions.json}. Reload with \code{mx_crypto_sessions_load()}.
#'
#' @param sessions A session set.
#' @param store_dir Character. Crypto store directory.
#' @return The path written, invisibly.
#' @examples
#' \donttest{
#' if (requireNamespace("mx.crypto", quietly = TRUE)) {
#'   dir <- file.path(tempfile(), "crypto")
#'   mx_crypto_sessions_save(mx_crypto_sessions_new(), dir)
#' }
#' }
#' @export
mx_crypto_sessions_save <- function(sessions, store_dir) {
    mx_require_crypto()
    dir.create(store_dir, showWarnings = FALSE, recursive = TRUE)
    key <- mx_crypto_key(store_dir)
    blob <- list(
                 olm = lapply(sessions$olm, function(s) {
        mx.crypto::mxc_olm_session_pickle(s, key)
    }),
                 olm_in = lapply(sessions$olm_in, function(s) {
        mx.crypto::mxc_olm_session_pickle(s, key)
    }),
                 megolm_out = lapply(sessions$megolm_out, function(m) {
        list(session = mx.crypto::mxc_megolm_outbound_pickle(m$session, key),
             shared = as.list(m$shared))
    }),
                 megolm_in = lapply(sessions$megolm_in, function(e) {
        list(session = mx.crypto::mxc_megolm_inbound_pickle(e$session, key),
             sender = e$sender %||% NA_character_,
             sender_ed25519 = e$sender_ed25519 %||% NA_character_,
             sender_bound = isTRUE(e$sender_bound))
    })
    )
    path <- file.path(store_dir, "sessions.json")
    writeLines(jsonlite::toJSON(blob, auto_unbox = TRUE), path)
    Sys.chmod(path, mode = "0600")
    invisible(path)
}

#' Load a session set from the crypto store
#'
#' @param store_dir Character. Crypto store directory.
#' @return A session set (empty if nothing is stored yet).
#' @examples
#' \donttest{
#' if (requireNamespace("mx.crypto", quietly = TRUE)) {
#'   s <- mx_crypto_sessions_load(file.path(tempfile(), "crypto"))
#' }
#' }
#' @export
mx_crypto_sessions_load <- function(store_dir) {
    mx_require_crypto()
    path <- file.path(store_dir, "sessions.json")
    if (!file.exists(path)) {
        return(mx_crypto_sessions_new())
    }
    key <- mx_crypto_key(store_dir)
    blob <- jsonlite::fromJSON(paste(readLines(path, warn = FALSE),
                                     collapse = "\n"),
                               simplifyVector = FALSE)
    out <- mx_crypto_sessions_new()
    for (nm in names(blob$olm %||% list())) {
        out$olm[[nm]] <- mx.crypto::mxc_olm_session_unpickle(blob$olm[[nm]], key)
    }
    for (nm in names(blob$olm_in %||% list())) {
        out$olm_in[[nm]] <- mx.crypto::mxc_olm_session_unpickle(
            blob$olm_in[[nm]], key)
    }
    for (nm in names(blob$megolm_out %||% list())) {
        m <- blob$megolm_out[[nm]]
        out$megolm_out[[nm]] <- list(
                                     session = mx.crypto::mxc_megolm_outbound_unpickle(m$session, key),
                                     shared = unlist(m$shared, use.names = FALSE) %||% character())
    }
    for (nm in names(blob$megolm_in %||% list())) {
        e <- blob$megolm_in[[nm]]
        # Stores written before sender attestation held a bare pickle
        # string here. Read them rather than discarding them: dropping the
        # entry would cost a running client its ability to decrypt every
        # message in that session's history. Such sessions carry no
        # attested sender, so events they decrypt report
        # sender_verified = FALSE.
        if (is.character(e) || !is.list(e)) {
            out$megolm_in[[nm]] <- list(
                                        session = mx.crypto::mxc_megolm_inbound_unpickle(e, key),
                                        sender = NA_character_,
                                        sender_ed25519 = NA_character_,
                                        sender_bound = FALSE)
            next
        }
        out$megolm_in[[nm]] <- list(
                                    session = mx.crypto::mxc_megolm_inbound_unpickle(e$session, key),
                                    sender = e$sender %||% NA_character_,
                                    sender_ed25519 = e$sender_ed25519 %||% NA_character_,
                                    sender_bound = isTRUE(e$sender_bound))
    }
    out
}

#' Encrypt an event for an encrypted room's devices
#'
#' Ensures an outbound Megolm session for the room, shares its key with
#' any recipient device that has not received it yet (establishing an Olm
#' session, claiming a one-time key when needed), and encrypts the event.
#' Returns the to-device payloads and the \code{m.room.encrypted} event;
#' the caller sends them with \code{mx.api::mx_send_to_device()} and
#' \code{mx.api::mx_send()}.
#'
#' @param account An mx.crypto account handle.
#' @param sessions A session set.
#' @param room_id Character room id.
#' @param content Named list. Plaintext event content.
#' @param sender_curve25519 Character. This device's Curve25519 key.
#' @param device_id Character. This device's id.
#' @param recipients List of recipient devices, each a list with
#'   \code{user_id}, \code{device_id}, \code{curve25519}, \code{ed25519},
#'   and (only needed to open a new Olm session) \code{otk}, a claimed
#'   one-time key. Use \code{mx_crypto_known_devices()}, which returns
#'   exactly this shape for devices whose keys verified.
#' @param sender_user_id Character. This user's Matrix id. Required to
#'   build a spec-conformant Olm payload that the recipient can attribute.
#' @return List with \code{to_device} (per-device payloads), \code{event}
#'   (the \code{m.room.encrypted} content), and the updated \code{sessions}.
#' @examples
#' \donttest{
#' if (requireNamespace("mx.crypto", quietly = TRUE)) {
#'   acct <- mx.crypto::mxc_account_new()
#'   out <- mx_crypto_encrypt_for_devices(
#'     acct, mx_crypto_sessions_new(), "!r:ex",
#'     list(msgtype = "m.text", body = "hi"),
#'     mx.crypto::mxc_account_identity_keys(acct)$curve25519, "DEV",
#'     recipients = list(), sender_user_id = "@me:ex")
#'   names(out)
#' }
#' }
#' @export
mx_crypto_encrypt_for_devices <- function(account, sessions, room_id,
    content, sender_curve25519,
    device_id, recipients = list(),
    sender_user_id = NULL) {
    mx_require_crypto()
    if (length(recipients) && is.null(sender_user_id)) {
        stop("sender_user_id is required to share room keys; without it the ",
             "Olm payload cannot name its sender and recipients will ",
             "reject it", call. = FALSE)
    }
    sender_ed25519 <- mx.crypto::mxc_account_identity_keys(account)$ed25519
    mo <- sessions$megolm_out[[room_id]]
    if (is.null(mo)) {
        mo <- list(session = mx.crypto::mxc_megolm_outbound_new(),
                   shared = character())
    }

    to_device <- list()
    for (r in recipients) {
        peer <- r$curve25519
        if (peer %in% mo$shared) {
            next
        }
        if (is.null(r$ed25519) || !nzchar(r$ed25519)) {
            stop("recipient ", r$user_id %||% "<unknown>", "/",
                 r$device_id %||% "<unknown>", " has no ed25519 key; it must ",
                 "come from a verified device (mx_crypto_known_devices())",
                 call. = FALSE)
        }
        olm <- sessions$olm[[peer]]
        if (is.null(olm)) {
            if (is.null(r$otk)) {
                stop("no Olm session for ", peer,
                     " and no one-time key supplied to open one",
                     call. = FALSE)
            }
            olm <- mx.crypto::mxc_olm_create_outbound(
                account, peer_curve25519 = peer, peer_otk = r$otk)
            sessions$olm[[peer]] <- olm
        }
        td <- mx_crypto_room_key_payload(olm, sender_curve25519, peer,
            room_id, mo$session, sender_user_id, sender_ed25519,
            r$user_id, r$ed25519)
        to_device[[length(to_device) + 1L]] <- list(
            user_id = r$user_id, device_id = r$device_id, content = td)
        mo$shared <- c(mo$shared, peer)
    }

    event <- mx_crypto_encrypt_event(mo$session, content, room_id,
                                     sender_curve25519, device_id)
    sessions$megolm_out[[room_id]] <- mo
    list(to_device = to_device, event = event, sessions = sessions)
}

#' Process a sync response: store room keys, decrypt room events
#'
#' Handles inbound to-device \code{m.room.encrypted} (Olm) messages,
#' storing any \code{m.room_key} as an inbound Megolm session, then
#' decrypts \code{m.room.encrypted} timeline events whose session is
#' known. Returns normalized text events in the same shape as
#' \code{mx_extract_text_events()}, plus the updated session set.
#'
#' @param account An mx.crypto account handle.
#' @param sessions A session set.
#' @param sync_resp Parsed \code{/sync} response.
#' @param self_curve25519 Character. This device's Curve25519 key.
#' @param devices List of verified devices from
#'   \code{mx_crypto_known_devices()}, or NULL. Room keys arrive over Olm
#'   carrying a claimed sender, and anyone who can reach this device can
#'   send one, so the claim is only worth something once it is matched
#'   against a device whose \code{device_keys} verified. Without this
#'   list decrypted events always report \code{sender_verified = FALSE}:
#'   they still decrypt, but nothing attests to who sent them.
#' @param self_id Character or NULL. This user's Matrix id, for
#'   \code{is_self} tagging.
#' @return List with \code{events} (decrypted, normalized) and the updated
#'   \code{sessions}.
#' @examples
#' \donttest{
#' if (requireNamespace("mx.crypto", quietly = TRUE)) {
#'   acct <- mx.crypto::mxc_account_new()
#'   res <- mx_crypto_process_sync(acct, mx_crypto_sessions_new(),
#'     list(to_device = list(events = list()), rooms = list(join = list())),
#'     mx.crypto::mxc_account_identity_keys(acct)$curve25519)
#'   length(res$events)
#' }
#' }
#' @export
mx_crypto_process_sync <- function(account, sessions, sync_resp,
                                   self_curve25519, self_id = NULL,
                                   devices = NULL) {
    mx_require_crypto()
    self_ed25519 <- mx.crypto::mxc_account_identity_keys(account)$ed25519

    # 1. To-device: recover shared room keys.
    for (ev in sync_resp$to_device$events %||% list()) {
        if (!isTRUE(ev$type == "m.room.encrypted") ||
            !isTRUE(ev$content$algorithm == MX_OLM)) {
            next
        }
        msg <- ev$content$ciphertext[[self_curve25519]]
        if (is.null(msg)) {
            next
        }
        sender <- ev$content$sender_key
        plaintext <- if (identical(as.integer(msg$type), 0L)) {
            res <- mx.crypto::mxc_olm_create_inbound(account,
                peer_curve25519 = sender, prekey_b64 = msg$body)
            sessions$olm_in[[sender]] <- res$session
            rawToChar(res$plaintext)
        } else {
            s <- sessions$olm_in[[sender]]
            if (is.null(s)) {
                next
            }
            rawToChar(mx.crypto::mxc_olm_decrypt(s, msg$type, msg$body))
        }
        decoded <- jsonlite::fromJSON(plaintext, simplifyVector = FALSE)
        chk <- mx_crypto_check_olm_payload(decoded, self_id, self_ed25519,
            sender, devices)
        if (!chk$ok) {
            next
        }
        if (identical(decoded$type, "m.room_key")) {
            c <- decoded$content
            key <- paste(c$room_id, c$session_id, sep = "|")
            # Record who the payload claimed to be from, and separately
            # whether that claim was tied to a verified device. Keeping the
            # unbound claim is still useful -- it catches an ordinary user
            # forging a sender, since the server stamps the real one on the
            # envelope -- but only sender_bound justifies calling a
            # decrypted event's sender verified.
            sessions$megolm_in[[key]] <- list(
                session = mx.crypto::mxc_megolm_inbound_new(c$session_key),
                sender = decoded$sender,
                sender_ed25519 = decoded$keys$ed25519,
                sender_bound = chk$bound)
        }
    }

    # 2. Room timelines: decrypt what we have keys for.
    events <- list()
    joined <- sync_resp$rooms$join %||% list()
    for (rid in names(joined)) {
        for (ev in joined[[rid]]$timeline$events %||% list()) {
            if (!isTRUE(ev$type == "m.room.encrypted") ||
                !isTRUE(ev$content$algorithm == MX_MEGOLM)) {
                next
            }
            key <- paste(rid, ev$content$session_id, sep = "|")
            entry <- sessions$megolm_in[[key]]
            if (is.null(entry)) {
                next
            }
            dec <- tryCatch(mx_crypto_decrypt_event(entry$session, ev$content),
                            error = function(e) NULL)
            if (is.null(dec)) {
                next
            }
            # The session was handed to us over Olm by whoever claimed the
            # identity recorded here. A disagreement with the envelope is a
            # forgery either way, so drop it. But agreement alone proves
            # nothing against a hostile homeserver, which writes both the
            # envelope and (via an injected to-device message) the claim.
            # Only a sender tied to a verified device counts as verified.
            attested <- entry$sender
            verified <- FALSE
            if (!is.null(attested) && !is.na(attested)) {
                if (!identical(attested, ev$sender)) {
                    warning("mx.client: dropping event ",
                            ev$event_id %||% "<no id>", " in ", rid,
                            ": envelope claims sender ",
                            ev$sender %||% "<missing>",
                            " but the Megolm session was shared by ",
                            attested, call. = FALSE)
                    next
                }
                verified <- isTRUE(entry$sender_bound)
            }
            ct <- dec$content
            events[[length(events) + 1L]] <- list(
                room_id = rid,
                event_id = ev$event_id,
                sender = ev$sender,
                sender_verified = verified,
                is_self = isTRUE(ev$sender == self_id),
                body = ct$body,
                msgtype = ct$msgtype,
                mentions = ct$`m.mentions`$user_ids
            )
        }
    }

    list(events = events, sessions = sessions)
}

# Two-party E2EE messaging loopback: Alice encrypts for an encrypted
# room, Bob recovers the key from the to-device payload and decrypts the
# timeline event -- entirely through mx.client's send/sync integration,
# with a persistence round-trip and an established-session second message.

library(tinytest)

if (!requireNamespace("mx.crypto", quietly = TRUE)) {
    exit_file("mx.crypto not available (needs a Rust toolchain)")
}
library(mx.client)

ROOM <- "!secret:example.org"

alice <- mx.crypto::mxc_account_new()
bob <- mx.crypto::mxc_account_new()
alice_curve <- mx.crypto::mxc_account_identity_keys(alice)$curve25519
bob_curve <- mx.crypto::mxc_account_identity_keys(bob)$curve25519
alice_ed <- mx.crypto::mxc_account_identity_keys(alice)$ed25519
bob_ed <- mx.crypto::mxc_account_identity_keys(bob)$ed25519
mx.crypto::mxc_account_generate_one_time_keys(bob, 2L)
bob_otk <- mx.crypto::mxc_account_one_time_keys(bob)[[1]]

a_store <- file.path(tempfile(), "a")
b_store <- file.path(tempfile(), "b")
a_sess <- mx_crypto_sessions_new()
b_sess <- mx_crypto_sessions_new()

# Helper: assemble a sync response for Bob from Alice's send output.
bob_sync <- function(out, event_id, with_to_device = TRUE) {
    td <- if (with_to_device && length(out$to_device)) {
        lapply(out$to_device, function(p) {
            list(type = "m.room.encrypted", sender = "@alice:example.org",
                 content = p$content)
        })
    } else list()
    list(
        to_device = list(events = td),
        rooms = list(join = stats::setNames(list(list(timeline = list(
            events = list(list(type = "m.room.encrypted",
                               event_id = event_id, sender = "@alice:example.org",
                               content = out$event))))), ROOM))
    )
}

# ---- message 1: fresh session (prekey Olm + Megolm key share) ----
recips <- list(list(user_id = "@bob:example.org", device_id = "BOBDEV",
                    curve25519 = bob_curve, ed25519 = bob_ed, otk = bob_otk))
out1 <- mx_crypto_encrypt_for_devices(
    alice, a_sess, ROOM, list(msgtype = "m.text", body = "first secret"),
    alice_curve, "ALICEDEV", recipients = recips,
    sender_user_id = "@alice:example.org")
a_sess <- out1$sessions
expect_equal(length(out1$to_device), 1L)         # key shared with Bob

# Bob's verified view of Alice's device, as /keys/query would yield it.
alice_devs <- list(list(user_id = "@alice:example.org", device_id = "ALICEDEV",
                        curve25519 = alice_curve, ed25519 = alice_ed))

res1 <- mx_crypto_process_sync(bob, b_sess, bob_sync(out1, "$1"),
                               bob_curve, self_id = "@bob:example.org",
                               devices = alice_devs)
b_sess <- res1$sessions
expect_equal(length(res1$events), 1L)
expect_equal(res1$events[[1]]$body, "first secret")   # decrypted
expect_false(res1$events[[1]]$is_self)
expect_true(res1$events[[1]]$sender_verified)         # bound to a verified device

# Olm one-time keys are single-use, so each scenario below needs its own
# recipient account and its own key share; a replayed prekey message
# cannot establish a second session.
probe <- function(event_id) {
    acct <- mx.crypto::mxc_account_new()
    idk <- mx.crypto::mxc_account_identity_keys(acct)
    mx.crypto::mxc_account_generate_one_time_keys(acct, 1L)
    otk <- mx.crypto::mxc_account_one_time_keys(acct)[[1]]
    out <- mx_crypto_encrypt_for_devices(
        alice, mx_crypto_sessions_new(), ROOM,
        list(msgtype = "m.text", body = "probe"), alice_curve, "ALICEDEV",
        recipients = list(list(user_id = "@bob:example.org",
                               device_id = "BOBDEV",
                               curve25519 = idk$curve25519,
                               ed25519 = idk$ed25519, otk = otk)),
        sender_user_id = "@alice:example.org")
    list(account = acct, curve = idk$curve25519,
         sync = bob_sync(out, event_id))
}

# Without a device list the traffic decrypts but claims nothing: the
# sender identity in an Olm payload is written by whoever holds the
# session, and anyone can open one to us.
p1 <- probe("$1b")
unbound <- suppressWarnings(
    mx_crypto_process_sync(p1$account, mx_crypto_sessions_new(), p1$sync,
                           p1$curve, self_id = "@bob:example.org"))
expect_equal(length(unbound$events), 1L)              # still decrypts
expect_false(unbound$events[[1]]$sender_verified)     # but attests nothing

# A device list whose curve25519 is not the one that sent the room key
# must not bind. This is the hostile-homeserver case: it injects a room
# key claiming Alice and stamps the envelope to agree, so the two halves
# corroborate each other and only the device binding catches it.
p2 <- probe("$1c")
wrong_curve <- list(list(user_id = "@alice:example.org",
                         device_id = "ALICEDEV",
                         curve25519 = p2$curve,   # not the sending device
                         ed25519 = alice_ed))
forged_attest <- suppressWarnings(
    mx_crypto_process_sync(p2$account, mx_crypto_sessions_new(), p2$sync,
                           p2$curve, self_id = "@bob:example.org",
                           devices = wrong_curve))
expect_equal(length(forged_attest$events), 1L)
expect_false(forged_attest$events[[1]]$sender_verified)

# The matching device does bind.
p3 <- probe("$1d")
ok <- mx_crypto_process_sync(p3$account, mx_crypto_sessions_new(), p3$sync,
                             p3$curve, self_id = "@bob:example.org",
                             devices = alice_devs)
expect_true(ok$events[[1]]$sender_verified)

# ---- persist both sides, reload from disk ----
mx_crypto_sessions_save(a_sess, a_store)
mx_crypto_sessions_save(b_sess, b_store)
a_sess <- mx_crypto_sessions_load(a_store)
b_sess <- mx_crypto_sessions_load(b_store)
expect_true(ROOM %in% names(a_sess$megolm_out))       # outbound survived
expect_true(bob_curve %in% a_sess$megolm_out[[ROOM]]$shared)
expect_equal(length(b_sess$megolm_in), 1L)            # inbound survived

# ---- message 2: established session, no re-share, decrypt from store ----
out2 <- mx_crypto_encrypt_for_devices(
    alice, a_sess, ROOM, list(msgtype = "m.text", body = "second secret"),
    alice_curve, "ALICEDEV", recipients = recips,
    sender_user_id = "@alice:example.org")
a_sess <- out2$sessions
expect_equal(length(out2$to_device), 0L)              # already shared

res2 <- mx_crypto_process_sync(bob, b_sess,
                               bob_sync(out2, "$2", with_to_device = FALSE),
                               bob_curve, self_id = "@bob:example.org")
expect_equal(length(res2$events), 1L)
expect_equal(res2$events[[1]]$body, "second secret")  # decrypted from reloaded state
expect_true(res2$events[[1]]$sender_verified)         # survives the store round-trip

# ---- a forged envelope sender is dropped, not reported ----
# The server rewrites `sender` on the timeline event. The Megolm session
# was shared by Alice over Olm, so the lie is detectable.
forged <- bob_sync(out2, "$3", with_to_device = FALSE)
forged$rooms$join[[ROOM]]$timeline$events[[1]]$sender <- "@mallory:example.org"
res3 <- suppressWarnings(
    mx_crypto_process_sync(bob, b_sess, forged, bob_curve,
                           self_id = "@bob:example.org"))
expect_equal(length(res3$events), 0L)

# ---- an Olm payload addressed to someone else is dropped ----
carol <- mx.crypto::mxc_account_new()
carol_curve <- mx.crypto::mxc_account_identity_keys(carol)$curve25519
carol_ed <- mx.crypto::mxc_account_identity_keys(carol)$ed25519
mx.crypto::mxc_account_generate_one_time_keys(bob, 2L)
bob_otk2 <- mx.crypto::mxc_account_one_time_keys(bob)[[2]]
# Alice shares a key naming Carol as recipient, but sends it to Bob.
misaddressed <- mx_crypto_encrypt_for_devices(
    alice, mx_crypto_sessions_new(), "!other:example.org",
    list(msgtype = "m.text", body = "not for bob"), alice_curve, "ALICEDEV",
    recipients = list(list(user_id = "@carol:example.org",
                           device_id = "CAROLDEV", curve25519 = bob_curve,
                           ed25519 = carol_ed, otk = bob_otk2)),
    sender_user_id = "@alice:example.org")
res4 <- suppressWarnings(
    mx_crypto_process_sync(bob, mx_crypto_sessions_new(),
                           bob_sync(misaddressed, "$4"), bob_curve,
                           self_id = "@bob:example.org"))
expect_equal(length(res4$sessions$megolm_in), 0L)     # key not installed

# ---- a legacy store (bare pickle) still loads ----
legacy_store <- file.path(tempfile(), "legacy")
dir.create(legacy_store, recursive = TRUE)
mx_crypto_sessions_save(b_sess, legacy_store)
raw <- jsonlite::fromJSON(paste(readLines(
    file.path(legacy_store, "sessions.json"), warn = FALSE), collapse = "\n"),
    simplifyVector = FALSE)
raw$megolm_in <- lapply(raw$megolm_in, function(e) e$session)  # old shape
writeLines(jsonlite::toJSON(raw, auto_unbox = TRUE),
           file.path(legacy_store, "sessions.json"))
reloaded <- mx_crypto_sessions_load(legacy_store)
expect_equal(length(reloaded$megolm_in), length(b_sess$megolm_in))
res5 <- mx_crypto_process_sync(bob, reloaded,
                               bob_sync(out2, "$5", with_to_device = FALSE),
                               bob_curve, self_id = "@bob:example.org")
expect_equal(length(res5$events), 1L)                 # history still decrypts
expect_false(res5$events[[1]]$sender_verified)        # but carries no attestation

# Conservative markdown-to-Matrix-HTML formatting.

mx_html_escape <- function(x) {
    x <- gsub("&", "&amp;", x, fixed = TRUE)
    x <- gsub("<", "&lt;", x, fixed = TRUE)
    x <- gsub(">", "&gt;", x, fixed = TRUE)
    x
}

mx_markdown_inline_html <- function(x) {
    x <- mx_html_escape(x)
    x <- gsub("`([^`]+)`", "<code>\\1</code>", x, perl = TRUE)
    x <- gsub("\\*\\*([^*]+)\\*\\*", "<strong>\\1</strong>", x, perl = TRUE)
    x <- gsub("\\b_([^_]+)_\\b", "<em>\\1</em>", x, perl = TRUE)
    x
}

mx_table_row_cells <- function(x) {
    x <- trimws(x)
    if (startsWith(x, "|")) {
        x <- substring(x, 2L)
    }
    if (endsWith(x, "|")) {
        x <- substring(x, 1L, nchar(x) - 1L)
    }
    trimws(strsplit(x, "|", fixed = TRUE)[[1]])
}

mx_is_table_separator <- function(x) {
    cells <- mx_table_row_cells(x)
    length(cells) > 0L && all(grepl("^:?-{3,}:?$", cells))
}

mx_is_table_row <- function(x) {
    grepl("\\|", x) && nzchar(trimws(x))
}

mx_table_align <- function(sep) {
    cells <- mx_table_row_cells(sep)
    vapply(cells, function(cell) {
        left <- startsWith(cell, ":")
        right <- endsWith(cell, ":")
        if (left && right) {
            "center"
        } else if (right) {
            "right"
        } else if (left) {
            "left"
        } else {
            NA_character_
        }
    }, character(1))
}

mx_table_html_cells <- function(cells, tag = "td", align = NULL) {
    n <- length(cells)
    if (is.null(align)) {
        align <- rep(NA_character_, n)
    }
    if (length(align) < n) {
        align <- c(align, rep(NA_character_, n - length(align)))
    }
    paste(vapply(seq_len(n), function(i) {
        attr <- if (!is.na(align[[i]]) && nzchar(align[[i]])) {
            sprintf(" align=\"%s\"", align[[i]])
        } else {
            ""
        }
        sprintf("<%s%s>%s</%s>", tag, attr,
                mx_markdown_inline_html(cells[[i]]), tag)
    }, character(1)), collapse = "")
}

mx_table_to_html <- function(rows) {
    header <- mx_table_row_cells(rows[[1]])
    align <- mx_table_align(rows[[2]])
    body <- rows[-c(1L, 2L)]
    html <- c("<table>", "<thead><tr>",
              mx_table_html_cells(header, "th", align), "</tr></thead>")
    if (length(body)) {
        body_html <- vapply(body, function(row) {
            cells <- mx_table_row_cells(row)
            # Pad or trim body rows to header width, matching common GFM behavior.
            if (length(cells) < length(header)) {
                cells <- c(cells, rep("", length(header) - length(cells)))
            } else if (length(cells) > length(header)) {
                cells <- cells[seq_along(header)]
            }
            paste0("<tr>", mx_table_html_cells(cells, "td", align), "</tr>")
        }, character(1))
        html <- c(html, "<tbody>", body_html, "</tbody>")
    }
    paste0(c(html, "</table>"), collapse = "")
}

#' Convert a conservative markdown subset to Matrix custom HTML
#'
#' Supports headings, bullets, numbered lists, fenced code blocks, inline
#' code, bold, simple underscore emphasis, and GitHub-style pipe tables.
#'
#' @param text Character markdown body.
#' @return Character HTML suitable for m.room.message formatted_body.
#' @examples
#' mx_markdown_to_html("# Status\n- built\n- checked\n\nShip `0.1.0` **soon**")
#' @export
mx_markdown_to_html <- function(text) {
    lines <- strsplit(text %||% "", "\n", fixed = TRUE)[[1]]
    out <- character()
    in_pre <- FALSE
    in_ul <- FALSE
    in_ol <- FALSE
    close_lists <- function() {
        z <- character()
        if (in_ul) {
            z <- c(z, "</ul>")
            in_ul <<- FALSE
        }
        if (in_ol) {
            z <- c(z, "</ol>")
            in_ol <<- FALSE
        }
        z
    }
    i <- 1L
    while (i <= length(lines)) {
        ln <- lines[[i]]
        if (grepl("^```", ln)) {
            if (in_pre) {
                out <- c(out, "</code></pre>")
                in_pre <- FALSE
            } else {
                out <- c(out, close_lists(), "<pre><code>")
                in_pre <- TRUE
            }
            i <- i + 1L
            next
        }
        if (in_pre) {
            out <- c(out, mx_html_escape(ln))
            i <- i + 1L
            next
        }
        if (i < length(lines) && mx_is_table_row(ln) &&
            mx_is_table_separator(lines[[i + 1L]])) {
            j <- i + 2L
            while (j <= length(lines) && mx_is_table_row(lines[[j]])) {
                j <- j + 1L
            }
            out <- c(out, close_lists(), mx_table_to_html(lines[i:(j - 1L)]))
            i <- j
            next
        }
        if (!nzchar(trimws(ln))) {
            out <- c(out, close_lists())
            i <- i + 1L
            next
        }
        h <- regexec("^(#{1,6})\\s+(.+)$", ln, perl = TRUE)
        hm <- regmatches(ln, h)[[1]]
        if (length(hm)) {
            out <- c(out, close_lists())
            lvl <- nchar(hm[[2]])
            body <- hm[[3]]
            out <- c(out, sprintf("<h%d>%s</h%d>", lvl,
                                  mx_markdown_inline_html(body), lvl))
            i <- i + 1L
            next
        }
        b <- regexec("^\\s*[-*]\\s+(.+)$", ln, perl = TRUE)
        bm <- regmatches(ln, b)[[1]]
        if (length(bm)) {
            if (in_ol) {
                out <- c(out, "</ol>")
                in_ol <- FALSE
            }
            if (!in_ul) {
                out <- c(out, "<ul>")
                in_ul <- TRUE
            }
            body <- bm[[2]]
            out <- c(out, sprintf("<li>%s</li>", mx_markdown_inline_html(body)))
            i <- i + 1L
            next
        }
        # Both "1." and "1)" are ordered-list markers; matching only the
        # dot silently renders "1)" input as a paragraph.
        o <- regexec("^\\s*[0-9]+[.)]\\s+(.+)$", ln, perl = TRUE)
        om <- regmatches(ln, o)[[1]]
        if (length(om)) {
            if (in_ul) {
                out <- c(out, "</ul>")
                in_ul <- FALSE
            }
            if (!in_ol) {
                out <- c(out, "<ol>")
                in_ol <- TRUE
            }
            body <- om[[2]]
            out <- c(out, sprintf("<li>%s</li>", mx_markdown_inline_html(body)))
            i <- i + 1L
            next
        }
        out <- c(out, close_lists(), sprintf("<p>%s</p>",
                mx_markdown_inline_html(ln)))
        i <- i + 1L
    }
    if (in_pre) {
        out <- c(out, "</code></pre>")
    }
    out <- c(out, close_lists())
    paste(out, collapse = "")
}

#' Turn textual @mentions into Matrix pills in formatted HTML
#'
#' @param html Character HTML (e.g. from \code{\link{mx_markdown_to_html}}).
#' @param user_ids Character Matrix user ids, such as
#'   \code{"@jorge:example.org"}.
#' @return HTML with textual \code{@localpart} occurrences replaced by
#'   matrix.to links. Unmatched user ids leave the HTML unchanged; they
#'   can still be placed in \code{m.mentions} by \code{mx_send_text()}.
#' @examples
#' mx_pill_mentions("<p>ping @jorge</p>", "@jorge:example.org")
#' @export
mx_pill_mentions <- function(html, user_ids) {
    if (!length(user_ids)) {
        return(html)
    }
    for (uid in unique(user_ids)) {
        local <- sub("^@([^:]+):.*$", "\\1", uid)
        esc <- gsub("([][{}()+*^$.|\\\\?])", "\\\\\\1", local, perl = TRUE)
        pill <- sprintf("<a href=\"https://matrix.to/#/%s\">%s</a>",
                        mx_html_escape(uid), mx_html_escape(local))
        html <- gsub(paste0("@", esc, "(:[A-Za-z0-9._-]+)?\\b"), pill, html,
                     perl = TRUE, ignore.case = TRUE)
    }
    html
}

library(tinytest)

html <- mx.client::mx_markdown_to_html(paste(c(
    "# Heading",
    "",
    "- one `code`",
    "- **two**",
    "",
    "```",
    "x < y & z",
    "```"
), collapse = "\n"))

expect_true(grepl("<h1>Heading</h1>", html, fixed = TRUE))
expect_true(grepl("<ul><li>one <code>code</code></li><li><strong>two</strong></li></ul>", html, fixed = TRUE))
expect_true(grepl("x &lt; y &amp; z", html, fixed = TRUE))

# mention pills
p <- mx.client::mx_pill_mentions("hey @Jorge, ping @jorge:cornball.ai too",
                                 "@jorge:cornball.ai")
pill <- "<a href=\"https://matrix.to/#/@jorge:cornball.ai\">jorge</a>"
expect_equal(p, paste0("hey ", pill, ", ping ", pill, " too"))

# localpart with a dot doesn't go regex-wild; unmentioned text untouched
p2 <- mx.client::mx_pill_mentions("@j.r rules, @jXr does not",
                                  "@j.r:cornball.ai")
expect_true(grepl("matrix.to/#/@j.r:cornball.ai", p2, fixed = TRUE))
expect_true(grepl("@jXr does not", p2, fixed = TRUE))

# no textual occurrence: html unchanged (m.mentions still notifies)
expect_equal(mx.client::mx_pill_mentions("no names here", "@tiny:cornball.ai"),
             "no names here")

# GitHub-style pipe tables become Matrix custom HTML tables.
tab <- mx.client::mx_markdown_to_html(paste(c(
    "| Area | OKF | pensar |",
    "|---|---|---|",
    "| Links | Markdown links | `[[wikilinks]]` |",
    "| Strictness | only **type** | title, type, source |"
), collapse = "\n"))
expect_true(grepl("<table>", tab, fixed = TRUE))
expect_true(grepl("<th>Area</th>", tab, fixed = TRUE))
expect_true(grepl("<td><code>[[wikilinks]]</code></td>", tab, fixed = TRUE))
expect_true(grepl("<strong>type</strong>", tab, fixed = TRUE))

# Known-good Matrix room shape from the Cornelius CRAN downloads report.
# This pins the exact format that rendered in-client: blank line before the
# pipe table, backticked package names, right-aligned numeric/date columns,
# and ordinary prose after the table.
known_good <- mx.client::mx_markdown_to_html(paste(c(
    "**CRAN downloads weekly** (16 packages, as of 2026-07-20)",
    "",
    "| package | last week | last month | total |",
    "|---|---:|---:|---:|",
    "| `tinyoauth` | 63 | 543 | 601 |",
    "| `saber` | 51 | 443 | 2,395 |",
    "",
    "* Partial period: package has not been on CRAN for the full week or month."
), collapse = "\n"))
expect_true(grepl("<table>", known_good, fixed = TRUE))
expect_true(grepl("<p><strong>CRAN downloads weekly</strong>", known_good,
                  fixed = TRUE))
expect_true(grepl("<td><code>tinyoauth</code></td>", known_good, fixed = TRUE))
expect_true(grepl("<td align=\"right\">2,395</td>", known_good, fixed = TRUE))
expect_true(grepl("Partial period", known_good, fixed = TRUE))

# Ordered lists accept both "1." and "1)" markers. There was no coverage
# here, which is how a rewrite that dropped ")" support went unnoticed.
expect_equal(mx.client::mx_markdown_to_html("1. first\n2. second"),
             "<ol><li>first</li><li>second</li></ol>")
expect_equal(mx.client::mx_markdown_to_html("1) first\n2) second"),
             "<ol><li>first</li><li>second</li></ol>")

# End-to-end encryption orchestration: ties mx.crypto's Olm/Megolm
# ratchets to the Matrix payload shapes and (via mx.api) the keys and
# to-device transport. mx.crypto does the cryptography; this file builds
# and consumes the m.room.encrypted / m.room_key envelopes around it.
#
# mx.crypto is a Suggests: every entry point guards with this helper, so
# the package installs and the non-encrypted paths work without a Rust
# toolchain.

MX_MEGOLM <- "m.megolm.v1.aes-sha2"
MX_OLM <- "m.olm.v1.curve25519-aes-sha2"

mx_require_crypto <- function() {
    if (!requireNamespace("mx.crypto", quietly = TRUE)) {
        stop("End-to-end encryption requires the 'mx.crypto' package. ",
             "Install it from CRAN before calling this again.", call. = FALSE)
    }
}

# ---- crypto store ----------------------------------------------------

#' Directory holding this client's encryption state
#'
#' The crypto store keeps the pickled Olm account, the 32-byte key that
#' encrypts those pickles at rest, and (later) per-peer Olm and per-room
#' Megolm sessions. It lives beside the JSON config under
#' \code{tools::R_user_dir()}.
#'
#' @param app Character. Application namespace.
#' @param path Character or NULL. Explicit store directory.
#' @return Character directory path.
#' @examples
#' mx_crypto_store_dir("myapp", path = tempfile())
#' @export
mx_crypto_store_dir <- function(app = "mx.client", path = NULL) {
    if (!is.null(path)) {
        return(path.expand(path))
    }
    file.path(tools::R_user_dir(app, "data"), "crypto")
}

# 32 random bytes for pickle encryption, persisted at rest (mode 0600).
# Created once per store; reused thereafter.
mx_crypto_key <- function(store_dir) {
    dir.create(store_dir, showWarnings = FALSE, recursive = TRUE)
    keyfile <- file.path(store_dir, "pickle.key")
    if (file.exists(keyfile)) {
        return(readBin(keyfile, "raw", n = 32L))
    }
    key <- mx_crypto_random_bytes(32L)
    writeBin(key, keyfile)
    Sys.chmod(keyfile, mode = "0600")
    key
}

mx_crypto_random_bytes <- function(n) {
    con <- tryCatch(file("/dev/urandom", "rb", raw = TRUE),
                    error = function(e) NULL)
    if (!is.null(con)) {
        on.exit(close(con))
        return(readBin(con, "raw", n = n))
    }
    # Portable fallback: mx.crypto-free hosts (e.g. Windows) get a weaker
    # source. Documented; the bots run on Linux with /dev/urandom.
    as.raw(sample.int(256L, n, replace = TRUE) - 1L)
}

#' Load or create this client's Olm account
#'
#' Unpickles \code{account.pickle} from the store, or mints a fresh
#' account and persists it. The account holds the device's long-lived
#' Curve25519/Ed25519 identity keys.
#'
#' @param store_dir Character. Crypto store directory.
#' @return An mx.crypto account handle.
#' @examples
#' \donttest{
#' if (requireNamespace("mx.crypto", quietly = TRUE)) {
#'   store <- mx_crypto_store_dir("myapp", path = tempfile())
#'   acct <- mx_crypto_account(store)
#'   unlink(store, recursive = TRUE)
#' }
#' }
#' @export
mx_crypto_account <- function(store_dir) {
    mx_require_crypto()
    dir.create(store_dir, showWarnings = FALSE, recursive = TRUE)
    key <- mx_crypto_key(store_dir)
    pfile <- file.path(store_dir, "account.pickle")
    if (file.exists(pfile)) {
        pickle <- paste(readLines(pfile, warn = FALSE), collapse = "")
        return(mx.crypto::mxc_account_unpickle(pickle, key))
    }
    acct <- mx.crypto::mxc_account_new()
    mx_crypto_account_save(acct, store_dir)
    acct
}

#' Persist an Olm account to the store
#'
#' @param account An mx.crypto account handle.
#' @param store_dir Character. Crypto store directory.
#' @return The pickle path, invisibly.
#' @examples
#' \donttest{
#' if (requireNamespace("mx.crypto", quietly = TRUE)) {
#'   store <- mx_crypto_store_dir("myapp", path = tempfile())
#'   acct <- mx_crypto_account(store)
#'   mx_crypto_account_save(acct, store)
#'   unlink(store, recursive = TRUE)
#' }
#' }
#' @export
mx_crypto_account_save <- function(account, store_dir) {
    mx_require_crypto()
    dir.create(store_dir, showWarnings = FALSE, recursive = TRUE)
    key <- mx_crypto_key(store_dir)
    pfile <- file.path(store_dir, "account.pickle")
    writeLines(mx.crypto::mxc_account_pickle(account, key), pfile)
    Sys.chmod(pfile, mode = "0600")
    invisible(pfile)
}

# ---- device keys (for /keys/upload) ----------------------------------

#' Build a signed device_keys object for upload
#'
#' Produces the \code{device_keys} structure \code{/keys/upload} expects:
#' the device's public identity keys plus an Ed25519 signature over their
#' canonical JSON. Hand the result to \code{mx.api::mx_keys_upload()}.
#'
#' @param account An mx.crypto account handle.
#' @param user_id Character. Full Matrix user id.
#' @param device_id Character. This device's id.
#' @return A named list ready to upload.
#' @examples
#' \donttest{
#' if (requireNamespace("mx.crypto", quietly = TRUE)) {
#'   store <- mx_crypto_store_dir("myapp", path = tempfile())
#'   acct <- mx_crypto_account(store)
#'   dk <- mx_crypto_device_keys(acct, "@bot:example.org", "DEVICEID")
#'   unlink(store, recursive = TRUE)
#' }
#' }
#' \dontrun{
#' # Uploading needs a live homeserver session:
#' mx.api::mx_keys_upload(session, device_keys = dk)
#' }
#' @export
mx_crypto_device_keys <- function(account, user_id, device_id) {
    mx_require_crypto()
    idk <- mx.crypto::mxc_account_identity_keys(account)
    body <- list(
                 user_id = user_id,
                 device_id = device_id,
                 algorithms = list(MX_OLM, MX_MEGOLM),
                 keys = stats::setNames(
                                        list(idk$curve25519, idk$ed25519),
                                        c(paste0("curve25519:", device_id), paste0("ed25519:", device_id))
        )
    )
    sig <- mx.crypto::mxc_account_sign(account, mx.api::mx_canonical_json(body))
    body$signatures <- stats::setNames(
                                       list(stats::setNames(list(sig), paste0("ed25519:", device_id))),
                                       user_id
    )
    body
}

# ---- Olm + room-key sharing (outbound) -------------------------------

#' Encrypt a Megolm room key to one device as a to-device payload
#'
#' Wraps the outbound Megolm session's key in an \code{m.room_key} event,
#' Olm-encrypts it to the recipient device, and returns the
#' \code{m.room.encrypted} to-device content to hand to
#' \code{mx.api::mx_send_to_device()}.
#'
#' @param olm_session An outbound Olm session
#'   (\code{mx.crypto::mxc_olm_create_outbound()}).
#' @param sender_curve25519 Character. This device's Curve25519 key.
#' @param recipient_curve25519 Character. Target device's Curve25519 key.
#' @param room_id Character. Room the key is for.
#' @param megolm_out An outbound Megolm session.
#' @param sender_user_id Character. This user's Matrix id.
#' @param sender_ed25519 Character. This device's Ed25519 key.
#' @param recipient_user_id Character. Target user's Matrix id.
#' @param recipient_ed25519 Character. Target device's Ed25519 key.
#' @return A named list: the to-device \code{m.room.encrypted} content.
#' @examples
#' \dontrun{
#' content <- mx_crypto_room_key_payload(olm, my_curve, their_curve,
#'                                       "!room:ex", megolm_out,
#'                                       "@me:ex", my_ed, "@them:ex", their_ed)
#' }
#' @export
mx_crypto_room_key_payload <- function(olm_session, sender_curve25519,
                                       recipient_curve25519, room_id,
                                       megolm_out, sender_user_id,
                                       sender_ed25519, recipient_user_id,
                                       recipient_ed25519) {
    mx_require_crypto()
    info <- mx.crypto::mxc_megolm_outbound_info(megolm_out)
    # The sender/recipient/keys block is not decoration: it is what lets
    # the receiving device attribute the key to us and confirm the message
    # was addressed to it. Omitting it (as this did before) leaves the
    # payload unauthenticatable and other clients reject it outright.
    room_key <- list(
                     type = "m.room_key",
                     content = list(algorithm = MX_MEGOLM, room_id = room_id,
                                    session_id = info$session_id,
                                    session_key = info$session_key),
                     sender = sender_user_id,
                     recipient = recipient_user_id,
                     recipient_keys = list(ed25519 = recipient_ed25519),
                     keys = list(ed25519 = sender_ed25519)
    )
    plaintext <- mx.api::mx_canonical_json(room_key)
    ct <- mx.crypto::mxc_olm_encrypt(olm_session, charToRaw(plaintext))
    list(
         algorithm = MX_OLM,
         sender_key = sender_curve25519,
         ciphertext = stats::setNames(
                                      list(list(type = ct$type, body = ct$body)),
                                      recipient_curve25519
        )
    )
}

# ---- inbound (receive room key + decrypt) ----------------------------

# Confirm a decrypted Olm payload was addressed to us.
#
# An Olm message decrypting successfully only proves it was encrypted to
# our Curve25519 key. It does not prove the sender meant it for us in this
# context: a homeserver can replay a payload it captured elsewhere. The
# spec's defence is the recipient block inside the plaintext, which is
# covered by the ratchet and so cannot be rewritten by the server.
#
# Returns list(ok, bound). ok = FALSE (with a warning) means discard the
# payload. bound = TRUE means the claimed sender identity was tied to a
# device whose keys we verified via /keys/query; only then may anything
# downstream be reported as sender-verified.
#
# self_id may be NULL, in which case the user-id half of the recipient
# check is skipped; callers that know their own id should always pass it.
mx_crypto_check_olm_payload <- function(decoded, self_id, self_ed25519,
                                        sender_curve25519 = NULL,
                                        devices = NULL) {
    if (!is.null(self_id) && !identical(decoded$recipient, self_id)) {
        warning("mx.client: dropping Olm payload addressed to ",
                decoded$recipient %||% "<missing>", ", not ", self_id,
                call. = FALSE)
        return(list(ok = FALSE, bound = FALSE))
    }
    if (!identical(decoded$recipient_keys$ed25519, self_ed25519)) {
        warning("mx.client: dropping Olm payload whose recipient_keys do ",
                "not match this device's Ed25519 key", call. = FALSE)
        return(list(ok = FALSE, bound = FALSE))
    }
    if (is.null(decoded$sender) || is.null(decoded$keys$ed25519)) {
        warning("mx.client: dropping Olm payload with no sender identity",
                call. = FALSE)
        return(list(ok = FALSE, bound = FALSE))
    }
    list(ok = TRUE,
         bound = mx_crypto_sender_bound(decoded, sender_curve25519, devices))
}

# Tie a payload's claimed sender to a verified device.
#
# The sender/keys block inside the plaintext is written by whoever holds
# the Olm session, and anyone can open one to us: our Curve25519 key and
# our one-time keys are both public. So the claim on its own is worth
# nothing. Comparing it against the cleartext envelope is no better,
# because a hostile homeserver writes both halves and can simply make
# them agree.
#
# The way out is the one thing the server cannot forge: a device_keys
# object carrying a valid self-signature, which is what
# mx_crypto_known_devices() returns. Requiring (sender, ed25519,
# curve25519) to match one of those as a triple binds the claim to a real
# device. Without a device list there is nothing to bind against, so the
# answer is FALSE and callers must not claim verification.
mx_crypto_sender_bound <- function(decoded, sender_curve25519, devices) {
    if (is.null(devices) || !length(devices) || is.null(sender_curve25519)) {
        return(FALSE)
    }
    for (d in devices) {
        if (identical(d$user_id, decoded$sender) &&
            identical(d$ed25519, decoded$keys$ed25519) &&
            identical(d$curve25519, sender_curve25519)) {
            return(TRUE)
        }
    }
    warning("mx.client: Olm payload claims sender ", decoded$sender,
            " but no verified device matches its identity keys; treating ",
            "it as unattributed", call. = FALSE)
    FALSE
}

#' Decrypt an inbound Olm to-device payload
#'
#' Accepts an \code{m.room.encrypted} to-device content addressed to this
#' device and returns the decrypted event. When it is an \code{m.room_key},
#' the caller builds an inbound Megolm session from
#' \code{content$session_key} with \code{mx_crypto_inbound_session()}.
#'
#' The decrypted payload is checked against this device's identity before
#' it is returned: a message that does not name us as recipient is
#' dropped, because decrypting successfully only proves it was encrypted
#' to our key, not that it was meant for us here.
#'
#' @param account An mx.crypto account handle.
#' @param my_curve25519 Character. This device's Curve25519 key.
#' @param content The to-device \code{m.room.encrypted} content.
#' @param self_id Character or NULL. This user's Matrix id. When NULL the
#'   recipient user-id check is skipped; pass it whenever it is known.
#' @param self_ed25519 Character or NULL. This device's Ed25519 key.
#'   Defaults to the account's own key.
#' @param devices List of verified devices from
#'   \code{mx_crypto_known_devices()}, or NULL. Used to tie the payload's
#'   claimed sender to a device whose keys were verified. The result
#'   carries \code{sender_bound}, which is FALSE when no list is supplied
#'   or nothing matches; an unbound sender identity is a claim, not a
#'   fact, because anyone can open an Olm session to this device.
#' @return The decrypted event (a parsed list) with a \code{sender_bound}
#'   flag, or NULL if it was not for us or failed the recipient checks.
#' @examples
#' \dontrun{
#' ev <- mx_crypto_handle_to_device(acct, my_curve, td_content,
#'                                  self_id = "@me:example.org",
#'                                  devices = mx_crypto_known_devices(cl, uid))
#' if (identical(ev$type, "m.room_key") && isTRUE(ev$sender_bound)) {
#'   inb <- mx_crypto_inbound_session(ev$content$session_key)
#' }
#' }
#' @export
mx_crypto_handle_to_device <- function(account, my_curve25519, content,
                                       self_id = NULL, self_ed25519 = NULL,
                                       devices = NULL) {
    mx_require_crypto()
    msg <- content$ciphertext[[my_curve25519]]
    if (is.null(msg)) {
        return(NULL)
    }
    if (is.null(self_ed25519)) {
        self_ed25519 <- mx.crypto::mxc_account_identity_keys(account)$ed25519
    }
    sender <- content$sender_key
    if (identical(as.integer(msg$type), 0L)) {
        res <- mx.crypto::mxc_olm_create_inbound(account,
            peer_curve25519 = sender, prekey_b64 = msg$body)
        plaintext <- rawToChar(res$plaintext)
    } else {
        stop("no established Olm session for a non-prekey to-device message",
             call. = FALSE)
    }
    decoded <- jsonlite::fromJSON(plaintext, simplifyVector = FALSE)
    chk <- mx_crypto_check_olm_payload(decoded, self_id, self_ed25519, sender,
                                       devices)
    if (!chk$ok) {
        return(NULL)
    }
    decoded$sender_bound <- chk$bound
    decoded
}

#' Build an inbound Megolm session from a shared room key
#'
#' @param session_key Character. The \code{session_key} from an
#'   \code{m.room_key} event.
#' @return An inbound Megolm session.
#' @examples
#' \dontrun{
#' inb <- mx_crypto_inbound_session(ev$content$session_key)
#' }
#' @export
mx_crypto_inbound_session <- function(session_key) {
    mx_require_crypto()
    mx.crypto::mxc_megolm_inbound_new(session_key)
}

# ---- event encrypt / decrypt -----------------------------------------

#' Encrypt event content for a room (Megolm)
#'
#' Returns the \code{m.room.encrypted} content to send as the event body.
#'
#' @param megolm_out An outbound Megolm session.
#' @param content Named list. The plaintext event content (e.g. an
#'   \code{m.room.message}).
#' @param room_id Character. Room id.
#' @param sender_curve25519 Character. This device's Curve25519 key.
#' @param device_id Character. This device's id.
#' @return A named list: \code{m.room.encrypted} content.
#' @examples
#' \dontrun{
#' enc <- mx_crypto_encrypt_event(megolm_out,
#'   list(msgtype = "m.text", body = "hi"), "!room:ex", my_curve, "DEV")
#' }
#' @export
mx_crypto_encrypt_event <- function(megolm_out, content, room_id,
                                    sender_curve25519, device_id) {
    mx_require_crypto()
    info <- mx.crypto::mxc_megolm_outbound_info(megolm_out)
    payload <- list(type = "m.room.message", room_id = room_id,
                    content = content)
    ct <- mx.crypto::mxc_megolm_encrypt(
                                        megolm_out, charToRaw(mx.api::mx_canonical_json(payload)))
    list(
         algorithm = MX_MEGOLM,
         sender_key = sender_curve25519,
         device_id = device_id,
         session_id = info$session_id,
         ciphertext = ct
    )
}

#' Decrypt an m.room.encrypted event (Megolm)
#'
#' @param inbound_session An inbound Megolm session for the event's
#'   \code{session_id}.
#' @param encrypted The \code{m.room.encrypted} event content.
#' @return The decrypted event payload (a parsed list).
#' @examples
#' \dontrun{
#' ev <- mx_crypto_decrypt_event(inb, encrypted_content)
#' ev$content$body
#' }
#' @export
mx_crypto_decrypt_event <- function(inbound_session, encrypted) {
    mx_require_crypto()
    res <- mx.crypto::mxc_megolm_decrypt(inbound_session, encrypted$ciphertext)
    jsonlite::fromJSON(rawToChar(res$plaintext), simplifyVector = FALSE)
}

# End-to-end encryption loopback: two accounts exchange an encrypted
# room message entirely in-process, through mx.client's own API. No
# homeserver, no network -- just the crypto orchestration.

library(tinytest)

if (!requireNamespace("mx.crypto", quietly = TRUE)) {
    exit_file("mx.crypto not available (needs a Rust toolchain)")
}
library(mx.client)

ROOM <- "!room:example.org"

# ---- accounts (Alice = sender, Bob = recipient) ----
alice <- mx.crypto::mxc_account_new()
bob <- mx.crypto::mxc_account_new()
alice_idk <- mx.crypto::mxc_account_identity_keys(alice)
bob_idk <- mx.crypto::mxc_account_identity_keys(bob)

# Bob publishes a one-time key Alice can claim.
mx.crypto::mxc_account_generate_one_time_keys(bob, 1L)
bob_otk <- mx.crypto::mxc_account_one_time_keys(bob)[[1]]

# ---- device_keys are signed and verifiable ----
dk <- mx_crypto_device_keys(alice, "@alice:example.org", "ALICEDEV")
expect_equal(dk$user_id, "@alice:example.org")
expect_true("curve25519:ALICEDEV" %in% names(dk$keys))
expect_true(!is.null(dk$signatures[["@alice:example.org"]][["ed25519:ALICEDEV"]]))
# the signature verifies against the device's own ed25519 key
verified <- mx.crypto::mxc_verify_device_keys(
    dk, "@alice:example.org", "ALICEDEV",
    required_algorithms = c("m.olm.v1.curve25519-aes-sha2",
                            "m.megolm.v1.aes-sha2"))
expect_true(isTRUE(verified) || is.list(verified))

# ---- Alice establishes an Olm session and a Megolm session ----
olm <- mx.crypto::mxc_olm_create_outbound(
    alice, peer_curve25519 = bob_idk$curve25519, peer_otk = bob_otk)
megolm_out <- mx.crypto::mxc_megolm_outbound_new()

# ---- Alice shares the room key with Bob (to-device payload) ----
td <- mx_crypto_room_key_payload(
    olm, sender_curve25519 = alice_idk$curve25519,
    recipient_curve25519 = bob_idk$curve25519,
    room_id = ROOM, megolm_out = megolm_out,
    sender_user_id = "@alice:example.org",
    sender_ed25519 = alice_idk$ed25519,
    recipient_user_id = "@bob:example.org",
    recipient_ed25519 = bob_idk$ed25519)
expect_equal(td$algorithm, "m.olm.v1.curve25519-aes-sha2")
expect_true(bob_idk$curve25519 %in% names(td$ciphertext))

# ---- Bob receives the to-device payload, recovers the room key ----
room_key_ev <- mx_crypto_handle_to_device(bob, bob_idk$curve25519, td,
                                          self_id = "@bob:example.org")
expect_equal(room_key_ev$type, "m.room_key")
expect_equal(room_key_ev$content$room_id, ROOM)
expect_equal(room_key_ev$sender, "@alice:example.org")   # attested sender
expect_equal(room_key_ev$keys$ed25519, alice_idk$ed25519)
inbound <- mx_crypto_inbound_session(room_key_ev$content$session_key)

# ---- Alice encrypts a room message; Bob decrypts it ----
enc <- mx_crypto_encrypt_event(
    megolm_out, content = list(msgtype = "m.text", body = "secret hello"),
    room_id = ROOM, sender_curve25519 = alice_idk$curve25519,
    device_id = "ALICEDEV")
expect_equal(enc$algorithm, "m.megolm.v1.aes-sha2")
expect_equal(enc$session_id, room_key_ev$content$session_id)

dec <- mx_crypto_decrypt_event(inbound, enc)
expect_equal(dec$content$body, "secret hello")          # the payoff
expect_equal(dec$room_id, ROOM)

# a non-recipient device gets NULL from the to-device payload
expect_null(mx_crypto_handle_to_device(bob, "wrong-curve-key", td))

# ---- account store: persist, reload, identity stable ----
store <- file.path(tempfile(), "crypto")
acct <- mx_crypto_account(store)                         # mints + saves
id1 <- mx.crypto::mxc_account_identity_keys(acct)$curve25519
acct2 <- mx_crypto_account(store)                        # reloads pickle
id2 <- mx.crypto::mxc_account_identity_keys(acct2)$curve25519
expect_equal(id1, id2)
expect_true(file.exists(file.path(store, "account.pickle")))
expect_equal(length(mx.client:::mx_crypto_key(store)), 32L)  # stable 0600 key

if (requireNamespace("tinytest", quietly = TRUE)) {
  Sys.setenv(R_USER_CACHE_DIR  = tempfile("mx.client_cache_"),
             R_USER_DATA_DIR   = tempfile("mx.client_data_"),
             R_USER_CONFIG_DIR = tempfile("mx.client_config_"))
  tinytest::test_package("mx.client")
}

# Matrix HTML table helpers.

mx_table_coerce <- function(x) {
    if (is.data.frame(x)) {
        return(x)
    }
    if (is.matrix(x)) {
        return(as.data.frame(x, stringsAsFactors = FALSE))
    }
    if (is.list(x)) {
        return(as.data.frame(x, stringsAsFactors = FALSE, check.names = FALSE))
    }
    stop("x must be a data.frame, matrix, or list", call. = FALSE)
}

mx_table_plain <- function(x, header = TRUE) {
    x <- mx_table_coerce(x)
    rows <- vapply(seq_len(nrow(x)), function(i) {
        paste(vapply(x[i,, drop = FALSE], as.character, character(1)),
              collapse = " | ")
    }, character(1))
    # Track the HTML: emitting column names in the fallback while the
    # formatted_body omits them shows different tables to different clients.
    if (isTRUE(header)) {
        rows <- c(paste(names(x), collapse = " | "), rows)
    }
    paste(rows, collapse = "\n")
}

#' Render tabular data as Matrix custom HTML
#'
#' Produces the conservative table shape rendered by Matrix clients such as
#' FluffyChat 2.6.0+: a bare \code{<table>} containing \code{<tr>},
#' \code{<th>}, and \code{<td>} nodes. No CSS, colspan, rowspan, or custom
#' attributes are emitted.
#'
#' @param x A data frame, matrix, or list coercible to a data frame.
#' @param header Logical. Include a header row using column names.
#' @return Character HTML suitable for Matrix \code{formatted_body}.
#' @examples
#' mx_table_html(data.frame(A = 1:2, B = c("x", "y")))
#' @export
mx_table_html <- function(x, header = TRUE) {
    x <- mx_table_coerce(x)
    cell <- function(tag, value) {
        sprintf("<%s>%s</%s>", tag, mx_html_escape(as.character(value)), tag)
    }
    rows <- character()
    if (isTRUE(header)) {
        rows <- c(rows, paste0("<tr>",
                               paste(vapply(names(x), cell, character(1), tag = "th"),
                                     collapse = ""),
                               "</tr>"))
    }
    if (nrow(x)) {
        rows <- c(rows, vapply(seq_len(nrow(x)), function(i) {
            vals <- vapply(x[i,, drop = FALSE], as.character, character(1))
            paste0("<tr>", paste(vapply(vals, cell, character(1), tag = "td"),
                                 collapse = ""), "</tr>")
        }, character(1)))
    }
    paste0("<table>", paste(rows, collapse = ""), "</table>")
}

#' Send tabular data to a Matrix room
#'
#' Sends a plain-text fallback body plus Matrix custom HTML table in
#' \code{formatted_body}. This bypasses Markdown entirely.
#'
#' @param client Matrix client config.
#' @param x A data frame, matrix, or list coercible to a data frame.
#' @param room Character room id/name or NULL for the default room.
#' @param header Logical. Include a header row using column names.
#' @param title Optional text prepended to the plain fallback body.
#' @param room_cache Optional room name-to-id cache.
#' @param dry_run Logical. Print instead of sending.
#' @return Event id, or NULL on dry-run.
#' @examples
#' client <- list(room_id = "!default:example.org")
#' mx_send_table(client, data.frame(A = 1, B = 2), dry_run = TRUE)
#' @export
mx_send_table <- function(client, x, room = NULL, header = TRUE,
                          title = NULL, room_cache = NULL, dry_run = FALSE) {
    html <- mx_table_html(x, header = header)
    body <- mx_table_plain(x, header = header)
    if (!is.null(title) && nzchar(title)) {
        body <- paste(as.character(title), body, sep = "\n")
    }
    if (isTRUE(dry_run)) {
        message("=== mx_send_table (dry-run) [", room %||% "default",
                "] ===\n", body, "\n--- formatted_body ---\n", html)
        return(invisible(NULL))
    }
    rid <- mx_resolve_room(client, room, room_cache = room_cache)
    mx.api::mx_send(mx_client_session(client), rid, body, msgtype = "m.text",
                    extra = list(format = "org.matrix.custom.html", formatted_body = html))
}

library(tinytest)

html <- mx.client::mx_table_html(data.frame(A = c("1", "<x>"), B = c("&", "two")))
expect_equal(html, paste0(
    "<table>",
    "<tr><th>A</th><th>B</th></tr>",
    "<tr><td>1</td><td>&amp;</td></tr>",
    "<tr><td>&lt;x&gt;</td><td>two</td></tr>",
    "</table>"))

expect_null(mx.client::mx_send_table(list(room_id = "!default:example.org"),
                                     data.frame(A = 1, B = 2),
                                     dry_run = TRUE))

# header = FALSE has to drop the header from both renderings, or the
# plain-text fallback shows a different table than formatted_body does.
d <- data.frame(A = c("1", "2"), B = c("x", "y"))
expect_equal(mx.client:::mx_table_plain(d, header = TRUE), "A | B\n1 | x\n2 | y")
expect_equal(mx.client:::mx_table_plain(d, header = FALSE), "1 | x\n2 | y")
expect_false(grepl("<th>", mx.client::mx_table_html(d, header = FALSE)))

# Session construction and first-time login.

#' Build an mx.api session from a client config
#'
#' @param client Named list or \code{"mx_client_config"} with \code{server},
#'   \code{token}, \code{user_id}, and \code{device_id}.
#' @return An \code{"mx_session"} from \pkg{mx.api}.
#' @examples
#' s <- mx_client_session(list(server = "https://matrix.example.org",
#'                             token = "syt_example",
#'                             user_id = "@bot:example.org",
#'                             device_id = "DEVICEID"))
#' class(s)
#' @export
mx_client_session <- function(client) {
    required <- c("server", "token", "user_id", "device_id")
    missing <- required[!vapply(required, function(x) {
        !is.null(client[[x]]) && length(client[[x]]) == 1L &&
        nzchar(client[[x]])
    }, logical(1))]
    if (length(missing)) {
        stop("client config missing fields: ",
             paste(missing, collapse = ", "), call. = FALSE)
    }
    mx.api::mx_session(
                       server = client$server,
                       token = client$token,
                       user_id = client$user_id,
                       device_id = client$device_id
    )
}

#' Configure and save a Matrix client
#'
#' Logs in with \pkg{mx.api}, joins or records the target room, and writes
#' a reusable local config. Extra fields are merged into the saved config
#' so applications can persist their own defaults without reimplementing
#' login.
#'
#' @param server Character. Homeserver base URL.
#' @param user Character. Matrix user localpart or full MXID.
#' @param password Character. Account password.
#' @param room Character. Room ID or alias to join.
#' @param app Character. Application namespace.
#' @param path Character or NULL. Explicit destination path.
#' @param device_id Character or NULL. Existing device id to reuse.
#' @param extra Named list. Additional fields to save.
#' @return Saved \code{"mx_client_config"}, invisibly.
#' @examples
#' \dontrun{
#' # Needs a live homeserver and account credentials.
#' mx_client_configure("https://matrix.example.org", "bot", "secret",
#'                     room = "#general:example.org", app = "myapp")
#' }
#' @export
mx_client_configure <- function(server, user, password, room,
                                app = "mx.client", path = NULL,
                                device_id = NULL, extra = list()) {
    if (!is.list(extra) || is.null(names(extra)) && length(extra)) {
        stop("extra must be a named list", call. = FALSE)
    }
    s <- mx.api::mx_login(server, user, password, device_id = device_id)
    room_id <- mx.api::mx_room_join(s, room)
    cfg <- list(server = server, user = user, password = password,
                token = s$token, user_id = s$user_id,
                device_id = s$device_id, room_id = room_id, sync_token = NULL)
    if (length(extra)) {
        cfg <- utils::modifyList(cfg, extra)
    }
    out <- mx_client_from_config(cfg, path = path, app = app)
    mx_client_save(out, app = app, path = path)
}

#' Re-login with stored credentials and refresh the saved token
#'
#' Uses the password persisted in the client config to obtain a fresh
#' access token for the \emph{same} device (reusing
#' \code{client$device_id}, so an E2EE device identity survives the
#' refresh), then saves the updated config. Typical use is recovering
#' from an invalidated token; see \code{\link{mx_with_relogin}} for the
#' catch-and-retry wrapper.
#'
#' @param client Matrix client config with \code{password}.
#' @param save Logical. Persist the refreshed config (default TRUE).
#' @return The refreshed \code{"mx_client_config"}.
#' @examples
#' \dontrun{
#' # Needs a live homeserver and a stored password.
#' client <- mx_client_relogin(mx_client_load("myapp"))
#' }
#' @export
mx_client_relogin <- function(client, save = TRUE) {
    if (is.null(client$password) || !nzchar(client$password)) {
        stop("client config has no stored password to re-login with",
             call. = FALSE)
    }
    s <- mx.api::mx_login(client$server, client$user, client$password,
                          device_id = client$device_id)
    refreshed <- mx_client_from_config(
                                       utils::modifyList(
            mx_client_plain_list(client),
            list(token = s$token, user_id = s$user_id, device_id = s$device_id)
        ),
                                       path = attr(client, "path"),
                                       app = attr(client, "app")
    )
    if (isTRUE(save)) {
        refreshed <- mx_client_save(refreshed)
    }
    refreshed
}

#' Run a client operation, re-logging in once on an expired token
#'
#' Calls \code{fn(client)}; if it fails with the server's invalid-token
#' error (\code{M_UNKNOWN_TOKEN}, signalled as a classed condition by
#' mx.api >= 0.3.0), re-logs in via \code{\link{mx_client_relogin}} and
#' retries once with the refreshed client. Any other error propagates.
#'
#' @param client Matrix client config with \code{password}.
#' @param fn Function taking a client config.
#' @param save Logical. Persist the refreshed config on relogin.
#' @return \code{fn}'s return value.
#' @examples
#' \dontrun{
#' mx_with_relogin(client, function(cl) {
#'     mx_send_text(cl, "still here after a token rotation")
#' })
#' }
#' @export
mx_with_relogin <- function(client, fn, save = TRUE) {
    tryCatch(
             fn(client),
             mx_error_M_UNKNOWN_TOKEN = function(e) {
        message("mx.client: token rejected; re-logging in")
        fn(mx_client_relogin(client, save = save))
    }
    )
}

<!--
%\VignetteEngine{simplermarkdown::mdweave_to_html}
%\VignetteIndexEntry{End-to-end encryption}
-->
---
title: "End-to-end encryption"
---

# End-to-end encryption

How to send and receive encrypted Matrix messages from R, and what each
step does on the wire. Code blocks here are display-only: the flow needs
a homeserver, a second device, and the `mx.crypto` package (which needs
a Rust toolchain to build).

`mx.crypto` provides the cryptographic primitives; `mx.client` keeps it
optional (Suggests) and only calls it from E2EE entry points, so
plaintext Matrix clients install and run without Rust.

## Security model

Read this first; it frames what the rest of the vignette delivers.

- **Trust on first use.** Device keys come from `/keys/query` and are
  used as-is. There is no cross-signing trust store yet, so a malicious
  homeserver could substitute device keys on first contact.
  (`mx.crypto` ships the verification primitives,
  `mxc_verify_device_keys()`; wiring a trust store is future work.)
- **No key requests or forwarded keys.** A device that missed the
  original `m.room_key` cannot ask peers to re-share it.
- **No SAS (emoji) verification.**
- **Local-key storage.** Ratchet state is pickled with a locally stored
  32-byte key (file mode 0600). That guards against casual inspection;
  it is only as strong as access to the local filesystem. No passphrase
  or hardware backing.

This matches what bots and controlled deployments need. For human-grade
verification flows, watch the `mx.crypto` roadmap.

## The pieces

Matrix encryption uses two ratchets, both provided by `mx.crypto`
(wrapping Matrix.org's `vodozemac`):

- **Olm** encrypts 1:1 between two *devices*. It is expensive to set up
  (a key agreement per peer device) and is used only to deliver secrets,
  chiefly Megolm session keys.
- **Megolm** encrypts room messages. One outbound session per room you
  send to; everyone you have shared its key with can decrypt.

So sending one encrypted room message means: have an Olm session with
each recipient device, send each of them your Megolm session key over
Olm (as *to-device* messages), then encrypt the actual message with
Megolm. mx.client orchestrates all of that; after one-time setup, a
send is one call and a receive is one call.

## Setup: a device with published keys

A Matrix *device* (a login session) gets long-lived identity keys plus a
pool of one-time keys others use to open Olm sessions with it. The
crypto store persists all of it (pickled with a locally stored 32-byte
key, mode 0600) so the device identity survives restarts.

```r
library(mx.client)

client <- mx_client_load(app = "myapp")    # see mx_client_configure()

store <- mx_crypto_store_dir("myapp")      # under tools::R_user_dir()
acct <- mx_crypto_account(store)           # load or mint identity keys
mx_crypto_publish_keys(client, acct, store, n_otks = 50L)
```

`mx_crypto_publish_keys()` builds the signed `device_keys` object,
signs and uploads one-time keys (`/keys/upload`), marks them published,
and saves the account. Run it again whenever the server's
`one_time_key_counts` runs low.

## Sending

```r
sessions <- mx_crypto_sessions_load(store)

res <- mx_send_encrypted(
    client, acct, sessions,
    room_id = "!abc:example.org",
    content = list(msgtype = "m.text", body = "the eagle lands at noon"),
    store_dir = store,
    member_ids = c("@friend:example.org")
)
res$event_id
```

On the wire, `mx_send_encrypted()`:

1. `/keys/query` — discovers the members' devices and identity keys
   (`mx_crypto_known_devices()`).
2. `/keys/claim` — claims a one-time key for each device we have no Olm
   session with (`mx_crypto_claim_otks()`), then opens the sessions.
3. `/sendToDevice` — Olm-encrypts the room's Megolm session key to each
   device that hasn't received it (`m.room_key` inside
   `m.room.encrypted`).
4. `/send` — Megolm-encrypts the actual content and posts the
   `m.room.encrypted` room event.
5. Persists the updated sessions to the store.

Steps 1–3 only do work the first time. Later sends to the same room are
a single Megolm encrypt + send.

## Receiving

```r
res <- mx_sync_update(client, timeout = 30000L)

my_curve <- mx.crypto::mxc_account_identity_keys(acct)$curve25519
out <- mx_crypto_process_sync(acct, sessions, res$sync, my_curve,
                              self_id = client$user_id)
sessions <- out$sessions
mx_crypto_sessions_save(sessions, store)

for (ev in out$events) cat(ev$sender, ":", ev$body, "\n")
```

`mx_crypto_process_sync()` makes two passes over the sync response:

1. **To-device events**: Olm-decrypts anything addressed to this
   device's Curve25519 key. Each recovered `m.room_key` becomes an
   inbound Megolm session, stored under `room_id|session_id`.
2. **Room timelines**: decrypts every `m.room.encrypted` event whose
   session is known, returning records in the same shape as
   `mx_extract_text_events()` — `room_id`, `event_id`, `sender`,
   `is_self`, `body`, `msgtype`, `mentions`.

Events whose keys haven't arrived yet are skipped, not errored; they
decrypt on a later pass once the to-device message lands.

## Persistence

Everything stateful lives in the crypto store directory:

| File | Holds |
|---|---|
| `pickle.key` | the locally stored 32-byte key the pickles are encrypted with |
| `account.pickle` | device identity (Curve25519 + Ed25519 keys, OTK state) |
| `sessions.json` | pickled Olm sessions and Megolm in/outbound sessions |

`mx_crypto_sessions_save()` / `mx_crypto_sessions_load()` round-trip the
session set, so an established room key keeps decrypting across process
restarts. One caution: the account binds to the config's `device_id`.
A homeserver will reject re-publishing different identity keys for an
existing device, so don't delete the store while keeping the login.

## Marking a room encrypted

Encryption is a room state event. Any member with permission can set it
(this is one-way; rooms don't downgrade to plaintext):

```r
s <- mx_client_session(client)
mx.api::mx_set_state(s, room_id, "m.room.encryption",
                     list(algorithm = "m.megolm.v1.aes-sha2"))
```

The boundaries of what this layer does are in the **Security model**
section at the top.

---
name: matrix-messaging
description: >
  Send and receive Matrix messages from R using the mx.* package family
  (mx.api / mx.crypto / mx.client). Use when a user wants an R program or
  agent to post to a Matrix room, read new messages, accept invites, send
  files or tables, or talk to a Matrix homeserver. Posts go through
  mx.client (config, room resolution, HTML formatting) over mx.api, never
  hand-rolled curl. End-to-end encryption is orchestrated over the optional
  mx.crypto package.
allowed-tools: Bash(r:*), Bash(Rscript:*), Read
---

# matrix-messaging

Drive a Matrix homeserver from R with the mx.* family. `mx.client` owns
config persistence, room resolution, sync cursors, event extraction, and
HTML formatting; it sits on `mx.api` (the raw Client-Server endpoints) and,
for encryption, `mx.crypto` (Olm/Megolm primitives).

Do not hand-roll curl against the homeserver. `mx.client` already owns
config loading, session construction, room resolution, and HTML formatting.

## The package family

| Package | Role | Notes |
|---|---|---|
| `mx.api` | HTTP transport, one function per CS-API endpoint, holds nothing | on CRAN; `mx.client` Imports it |
| `mx.crypto` | Olm/Megolm primitives (vodozemac via Rust) | on CRAN; `mx.client` Suggests it (E2EE only) |
| `mx.client` | stateful client: config, rooms, sync, formatting, E2EE orchestration | on CRAN |

## First-time setup

`mx_client_configure()` logs in, joins the room, and persists credentials
(mode 0600, under `tools::R_user_dir()`). `app` namespaces the stored
config so several bots can coexist. Do this once per identity.

```r
mx.client::mx_client_configure(
    server   = "https://matrix.example.org",
    user     = "bot",
    password = "secret",
    room     = "#general:example.org",   # default room for later sends
    app      = "myapp"
)
```

## Send a message

Every later session loads the stored config and sends. `markdown = TRUE`
adds a conservative HTML `formatted_body` (headings, bold, code, lists,
links, and GitHub pipe tables become Matrix `<table>` HTML). `room` takes a
name or id; omit it for the configured default. `mx_send_text()` returns the
event id.

```r
client <- mx.client::mx_client_load(app = "myapp")

mx.client::mx_send_text(client, "hello from R")                  # default room
id <- mx.client::mx_send_text(client, "**shipped** `v0.1.1`",
                              room = "general", markdown = TRUE)  # named room
```

Tokens stay in the config file. Never print them.

To load a config by explicit path instead of `app`, pass `path =`. Use
`dry_run = TRUE` to print the resolved send without hitting the network.

### Mentions

`mentions = "@user:example.org"` adds the user to the event's `m.mentions`
(so they get pinged) and rewrites any textual `@localpart` in the body into a
matrix.to pill. A pill implies an HTML body even without `markdown = TRUE`.

## Resolve rooms

Send by human name instead of `!opaque:id`. `mx_resolve_room()` turns a name
into an id (or passes a literal `!id`/`#alias` through); `mx_room_lookup_by_name()`
lists the joined rooms, which is how you find a DM (the room whose members
are just the bot and one person).

```r
room_id <- mx.client::mx_resolve_room(client, "general")
mx.client::mx_room_lookup_by_name(client)        # name -> id table
```

## Send files and media

`mx_send_media()` (needs `mx.api (>= 0.3.0)`) uploads and posts in one call.
The msgtype comes from the file's MIME type: a `.png` posts as `m.image`, a
`.mp4` as `m.video`. Server upload cap is about 20 MB
(`mx.api::mx_media_config()` to check). Pass `content_type =` for files whose
extension lies, and `body =` for a caption/filename.

```r
mx.client::mx_send_media(client, "plot.png", room = "general")
```

## Send a table

`mx_send_table()` renders a data frame (or matrix) straight to Matrix HTML
via `mx_table_html()`.

```r
mx.client::mx_send_table(client, head(mtcars), room = "general")
```

## Receive: read new messages and advance the cursor

`mx_sync_update()` long-polls and advances the stored sync cursor so you only
see new events. The `mx_extract_*` helpers parse the sync response.

```r
res    <- mx.client::mx_sync_update(client, timeout = 30000L)
msgs   <- mx.client::mx_extract_text_events(res$sync, client$user_id)
invs   <- mx.client::mx_extract_invites(res$sync)
mx.client::mx_accept_invites(client, invs)        # join rooms you were invited to
```

## Survive token rotation

`mx_with_relogin()` wraps any client operation: on `M_UNKNOWN_TOKEN` it
re-logs in with the stored password (keeping the device id, so an E2EE
identity survives), saves the refreshed token, and retries once.

```r
mx.client::mx_with_relogin(client, function(cl) {
    mx.client::mx_send_text(cl, "still here after a token rotation")
})
```

## End-to-end encryption

Olm/Megolm send/receive orchestrated over `mx.crypto`, aimed at bots and
controlled deployments. `mx.crypto` is a Suggests and is only touched from
the E2EE entry points, so plaintext clients install and run without a Rust
toolchain. Security model is trust-on-first-use (no cross-signing trust store
yet, no key-request flow). Check a room's state with `mx_room_encrypted()`
before choosing the encrypted or plaintext path.

The full flow (store, account, key publish, `mx_send_encrypted()`,
`mx_crypto_process_sync()`) and its current limitations are in
`vignette("e2ee", package = "mx.client")`.

## Report

End with: which identity (app/config) posted, which room, the message, and
the returned `event_id`.

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_account_save}
\alias{mx_crypto_account_save}
\title{Persist an Olm account to the store}
\usage{
mx_crypto_account_save(account, store_dir)
}
\arguments{
\item{account}{An mx.crypto account handle.}

\item{store_dir}{Character. Crypto store directory.}
}
\value{
The pickle path, invisibly.
}
\description{
Persist an Olm account to the store
}
\examples{
\donttest{
if (requireNamespace("mx.crypto", quietly = TRUE)) {
  store <- mx_crypto_store_dir("myapp", path = tempfile())
  acct <- mx_crypto_account(store)
  mx_crypto_account_save(acct, store)
  unlink(store, recursive = TRUE)
}
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_claim_otks}
\alias{mx_crypto_claim_otks}
\title{Claim a one-time key for each device}
\usage{
mx_crypto_claim_otks(client, devices)
}
\arguments{
\item{client}{Matrix client config.}

\item{devices}{List of devices from \code{mx_crypto_known_devices()}.}
}
\value{
The devices with an \code{otk} field added where one was
  claimed and verified.
}
\description{
Calls \code{/keys/claim} and attaches the claimed key to each device as
\code{$otk}, ready for \code{mx_crypto_encrypt_for_devices()}.
}
\details{
Each claimed key's signature is checked against the device's Ed25519
key, which must itself have come from a verified \code{device_keys}
(that is, from \code{mx_crypto_known_devices()}). A key that fails is
dropped with a warning and its device comes back with no \code{otk}.

}
\examples{
\dontrun{
devs <- mx_crypto_claim_otks(client, mx_crypto_known_devices(client, uid))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_sync_update}
\alias{mx_sync_update}
\title{Sync once and update the stored cursor}
\usage{
mx_sync_update(client, timeout = 0L, filter = NULL, save = TRUE, path = NULL,
               app = NULL)
}
\arguments{
\item{client}{Matrix client config.}

\item{timeout}{Integer long-poll timeout in milliseconds.}

\item{filter}{Character or NULL. Matrix sync filter.}

\item{save}{Logical. Persist the updated client config.}

\item{path}{Character or NULL. Save destination.}

\item{app}{Character or NULL. Application namespace for default saves.}
}
\value{
List with \code{sync}, \code{client}, and \code{first_run}.
}
\description{
Calls \code{mx.api::mx_sync()} using \code{client$sync_token}, stores
the returned \code{next_batch} in a returned client object, and
optionally saves it back to disk.
}
\examples{
\dontrun{
# Needs a live homeserver session.
client <- mx_client_load("myapp")
res <- mx_sync_update(client, timeout = 30000L)
events <- mx_extract_text_events(res$sync, client$user_id)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_sessions_new}
\alias{mx_crypto_sessions_new}
\title{Create an empty E2EE session set}
\usage{
mx_crypto_sessions_new()
}
\value{
A session set: named lists \code{olm}, \code{olm_in},
  \code{megolm_out}, \code{megolm_in}.
}
\description{
Create an empty E2EE session set
}
\examples{
s <- mx_crypto_sessions_new()
names(s)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_client_relogin}
\alias{mx_client_relogin}
\title{Re-login with stored credentials and refresh the saved token}
\usage{
mx_client_relogin(client, save = TRUE)
}
\arguments{
\item{client}{Matrix client config with \code{password}.}

\item{save}{Logical. Persist the refreshed config (default TRUE).}
}
\value{
The refreshed \code{"mx_client_config"}.
}
\description{
Uses the password persisted in the client config to obtain a fresh
access token for the \emph{same} device (reusing
\code{client$device_id}, so an E2EE device identity survives the
refresh), then saves the updated config. Typical use is recovering
from an invalidated token; see \code{\link{mx_with_relogin}} for the
catch-and-retry wrapper.
}
\examples{
\dontrun{
# Needs a live homeserver and a stored password.
client <- mx_client_relogin(mx_client_load("myapp"))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_decrypt_event}
\alias{mx_crypto_decrypt_event}
\title{Decrypt an m.room.encrypted event (Megolm)}
\usage{
mx_crypto_decrypt_event(inbound_session, encrypted)
}
\arguments{
\item{inbound_session}{An inbound Megolm session for the event's
\code{session_id}.}

\item{encrypted}{The \code{m.room.encrypted} event content.}
}
\value{
The decrypted event payload (a parsed list).
}
\description{
Decrypt an m.room.encrypted event (Megolm)
}
\examples{
\dontrun{
ev <- mx_crypto_decrypt_event(inb, encrypted_content)
ev$content$body
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_set_displayname}
\alias{mx_set_displayname}
\title{Set the account's profile display name}
\usage{
mx_set_displayname(client, name, save = TRUE)
}
\arguments{
\item{client}{Matrix client config.}

\item{name}{Character. New display name.}

\item{save}{Logical. On a relogin, persist the refreshed token to the
client's config path (see \code{\link{mx_client_relogin}}).}
}
\value{
TRUE, invisibly.
}
\description{
Client-level wrapper over \code{mx.api::mx_set_displayname()}: builds
the session from the client config and retries once through
\code{\link{mx_with_relogin}} when the homeserver rejects a rotated
token. The display name is account-global and shows in the sender
line of every message the account posts.
}
\examples{
\dontrun{
# A real rename needs a live homeserver session:
client <- mx_client_load("myapp")
mx_set_displayname(client, "mybot (maintenance)")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_store_dir}
\alias{mx_crypto_store_dir}
\title{Directory holding this client's encryption state}
\usage{
mx_crypto_store_dir(app = "mx.client", path = NULL)
}
\arguments{
\item{app}{Character. Application namespace.}

\item{path}{Character or NULL. Explicit store directory.}
}
\value{
Character directory path.
}
\description{
The crypto store keeps the pickled Olm account, the 32-byte key that
encrypts those pickles at rest, and (later) per-peer Olm and per-room
Megolm sessions. It lives beside the JSON config under
\code{tools::R_user_dir()}.
}
\examples{
mx_crypto_store_dir("myapp", path = tempfile())
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_accept_invites}
\alias{mx_accept_invites}
\title{Accept pending Matrix room invites}
\usage{
mx_accept_invites(client, invites)
}
\arguments{
\item{client}{Matrix client config.}

\item{invites}{Character vector of room ids.}
}
\value{
Character vector of joined room ids.
}
\description{
Accept pending Matrix room invites
}
\examples{
\dontrun{
# Needs a live homeserver session.
client <- mx_client_load("myapp")
res <- mx_sync_update(client)
mx_accept_invites(res$client, mx_extract_invites(res$sync))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_device_keys}
\alias{mx_crypto_device_keys}
\title{Build a signed device_keys object for upload}
\usage{
mx_crypto_device_keys(account, user_id, device_id)
}
\arguments{
\item{account}{An mx.crypto account handle.}

\item{user_id}{Character. Full Matrix user id.}

\item{device_id}{Character. This device's id.}
}
\value{
A named list ready to upload.
}
\description{
Produces the \code{device_keys} structure \code{/keys/upload} expects:
the device's public identity keys plus an Ed25519 signature over their
canonical JSON. Hand the result to \code{mx.api::mx_keys_upload()}.
}
\examples{
\donttest{
if (requireNamespace("mx.crypto", quietly = TRUE)) {
  store <- mx_crypto_store_dir("myapp", path = tempfile())
  acct <- mx_crypto_account(store)
  dk <- mx_crypto_device_keys(acct, "@bot:example.org", "DEVICEID")
  unlink(store, recursive = TRUE)
}
}
\dontrun{
# Uploading needs a live homeserver session:
mx.api::mx_keys_upload(session, device_keys = dk)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_client_from_config}
\alias{mx_client_from_config}
\title{Wrap a list as an mx.client config}
\usage{
mx_client_from_config(cfg, path = NULL, app = NULL)
}
\arguments{
\item{cfg}{Named list.}

\item{path}{Character or NULL. Source/sink path for saves.}

\item{app}{Character or NULL. Application namespace.}
}
\value{
An object of class \code{"mx_client_config"}.
}
\description{
Wrap a list as an mx.client config
}
\examples{
cfg <- mx_client_from_config(list(server = "https://matrix.example.org",
                                  token = "syt_example",
                                  user_id = "@bot:example.org",
                                  device_id = "DEVICEID"))
class(cfg)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_client_save}
\alias{mx_client_save}
\title{Save a Matrix client config}
\usage{
mx_client_save(client, app = NULL, path = NULL)
}
\arguments{
\item{client}{Named list or \code{"mx_client_config"}.}

\item{app}{Character or NULL. Application namespace.}

\item{path}{Character or NULL. Destination path.}
}
\value{
The saved config, invisibly.
}
\description{
Writes JSON with mode 0600.
}
\examples{
path <- file.path(tempdir(), "matrix.json")
mx_client_save(list(server = "https://matrix.example.org",
                    token = "syt_example",
                    user_id = "@bot:example.org",
                    device_id = "DEVICEID"),
               path = path)
unlink(path)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_encrypt_for_devices}
\alias{mx_crypto_encrypt_for_devices}
\title{Encrypt an event for an encrypted room's devices}
\usage{
mx_crypto_encrypt_for_devices(account, sessions, room_id, content,
                              sender_curve25519, device_id,
                              recipients = list(), sender_user_id = NULL)
}
\arguments{
\item{account}{An mx.crypto account handle.}

\item{sessions}{A session set.}

\item{room_id}{Character room id.}

\item{content}{Named list. Plaintext event content.}

\item{sender_curve25519}{Character. This device's Curve25519 key.}

\item{device_id}{Character. This device's id.}

\item{recipients}{List of recipient devices, each a list with
\code{user_id}, \code{device_id}, \code{curve25519}, \code{ed25519},
and (only needed to open a new Olm session) \code{otk}, a claimed
one-time key. Use \code{mx_crypto_known_devices()}, which returns
exactly this shape for devices whose keys verified.}

\item{sender_user_id}{Character. This user's Matrix id. Required to
build a spec-conformant Olm payload that the recipient can attribute.}
}
\value{
List with \code{to_device} (per-device payloads), \code{event}
  (the \code{m.room.encrypted} content), and the updated \code{sessions}.
}
\description{
Ensures an outbound Megolm session for the room, shares its key with
any recipient device that has not received it yet (establishing an Olm
session, claiming a one-time key when needed), and encrypts the event.
Returns the to-device payloads and the \code{m.room.encrypted} event;
the caller sends them with \code{mx.api::mx_send_to_device()} and
\code{mx.api::mx_send()}.
}
\examples{
\donttest{
if (requireNamespace("mx.crypto", quietly = TRUE)) {
  acct <- mx.crypto::mxc_account_new()
  out <- mx_crypto_encrypt_for_devices(
    acct, mx_crypto_sessions_new(), "!r:ex",
    list(msgtype = "m.text", body = "hi"),
    mx.crypto::mxc_account_identity_keys(acct)$curve25519, "DEV",
    recipients = list(), sender_user_id = "@me:ex")
  names(out)
}
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_markdown_to_html}
\alias{mx_markdown_to_html}
\title{Convert a conservative markdown subset to Matrix custom HTML}
\usage{
mx_markdown_to_html(text)
}
\arguments{
\item{text}{Character markdown body.}
}
\value{
Character HTML suitable for m.room.message formatted_body.
}
\description{
Supports headings, bullets, numbered lists, fenced code blocks, inline
code, bold, simple underscore emphasis, and GitHub-style pipe tables.
}
\examples{
mx_markdown_to_html("# Status\n- built\n- checked\n\nShip `0.1.0` **soon**")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_extract_reaction_verdict}
\alias{mx_extract_reaction_verdict}
\title{Extract a reaction approval verdict from sync events}
\usage{
mx_extract_reaction_verdict(sync_resp, room_id, self_id, target_event_id,
                            approve_keys = NULL, deny_keys = NULL)
}
\arguments{
\item{sync_resp}{Parsed \code{/sync} response.}

\item{room_id}{Character room id.}

\item{self_id}{Current user's Matrix id.}

\item{target_event_id}{Event id being reacted to.}

\item{approve_keys}{Character vector of reaction keys read as approval.
\code{NULL} (default) uses thumbs-up (U+1F44D), check-mark (U+2705),
and \code{"y"}/\code{"yes"}/\code{"ok"}.}

\item{deny_keys}{Character vector of reaction keys read as denial.
\code{NULL} (default) uses thumbs-down (U+1F44E), cross-mark
(U+274C), and \code{"n"}/\code{"no"}/\code{"nope"}.}
}
\value{
TRUE, FALSE, or NULL.
}
\description{
Scans a room timeline for a reaction on \code{target_event_id} from
someone other than \code{self_id}. Returns TRUE for approval keys,
FALSE for denial keys, or NULL when no verdict is present.
}
\examples{
sync_resp <- list(rooms = list(join = list("!room:example.org" = list(
    timeline = list(events = list(list(type = "m.reaction",
        sender = "@alice:example.org",
        content = list("m.relates_to" = list(rel_type = "m.annotation",
            event_id = "$msg", key = "yes")))))))))
mx_extract_reaction_verdict(sync_resp, "!room:example.org",
                            self_id = "@bot:example.org",
                            target_event_id = "$msg")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_send_table}
\alias{mx_send_table}
\title{Send tabular data to a Matrix room}
\usage{
mx_send_table(client, x, room = NULL, header = TRUE, title = NULL,
              room_cache = NULL, dry_run = FALSE)
}
\arguments{
\item{client}{Matrix client config.}

\item{x}{A data frame, matrix, or list coercible to a data frame.}

\item{room}{Character room id/name or NULL for the default room.}

\item{header}{Logical. Include a header row using column names.}

\item{title}{Optional text prepended to the plain fallback body.}

\item{room_cache}{Optional room name-to-id cache.}

\item{dry_run}{Logical. Print instead of sending.}
}
\value{
Event id, or NULL on dry-run.
}
\description{
Sends a plain-text fallback body plus Matrix custom HTML table in
\code{formatted_body}. This bypasses Markdown entirely.
}
\examples{
client <- list(room_id = "!default:example.org")
mx_send_table(client, data.frame(A = 1, B = 2), dry_run = TRUE)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_known_devices}
\alias{mx_crypto_known_devices}
\title{List the devices (and identity keys) of some users}
\usage{
mx_crypto_known_devices(client, user_ids)
}
\arguments{
\item{client}{Matrix client config.}

\item{user_ids}{Character vector of Matrix user ids.}
}
\value{
List of verified devices, each \code{list(user_id, device_id,
  curve25519, ed25519)}.
}
\description{
Queries \code{/keys/query} and flattens the result to a list of devices.
}
\details{
Devices are returned only if their Ed25519 self-signature verifies
against the identity the homeserver claims for them. A device that
fails is dropped with a warning naming it, and the remaining devices
are still returned: one bad device must not make a room unusable.

}
\examples{
\dontrun{
mx_crypto_known_devices(client, "@bob:example.org")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_client_session}
\alias{mx_client_session}
\title{Build an mx.api session from a client config}
\usage{
mx_client_session(client)
}
\arguments{
\item{client}{Named list or \code{"mx_client_config"} with \code{server},
\code{token}, \code{user_id}, and \code{device_id}.}
}
\value{
An \code{"mx_session"} from \pkg{mx.api}.
}
\description{
Build an mx.api session from a client config
}
\examples{
s <- mx_client_session(list(server = "https://matrix.example.org",
                            token = "syt_example",
                            user_id = "@bot:example.org",
                            device_id = "DEVICEID"))
class(s)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_room_lookup_by_name}
\alias{mx_room_lookup_by_name}
\title{Look up a joined room by display name}
\usage{
mx_room_lookup_by_name(client, name)
}
\arguments{
\item{client}{Matrix client config.}

\item{name}{Character room name.}
}
\value{
Room id, or NULL when no joined room has that name.
}
\description{
Look up a joined room by display name
}
\examples{
\dontrun{
# Needs a live homeserver session.
mx_room_lookup_by_name(mx_client_load("myapp"), "general")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_send_text}
\alias{mx_send_text}
\title{Send plain text to a Matrix room}
\usage{
mx_send_text(client, text, room = NULL, msgtype = "m.text", room_cache = NULL,
             dry_run = FALSE, markdown = FALSE, mentions = NULL)
}
\arguments{
\item{client}{Matrix client config.}

\item{text}{Character message body.}

\item{room}{Character room id/name or NULL for the default room.}

\item{msgtype}{Character Matrix message type.}

\item{room_cache}{Optional room name-to-id cache.}

\item{dry_run}{Logical. Print instead of sending.}

\item{markdown}{Logical. If TRUE, include Matrix custom HTML derived
from a conservative markdown subset.}

\item{mentions}{Character vector of Matrix user ids to mention
(e.g. \code{"@jorge:cornball.ai"}). Each id is added to the event's
\code{m.mentions} (so the user is notified) and any textual
\code{@localpart} in the body becomes a \code{matrix.to} pill in the
HTML. Implies an HTML formatted body even when \code{markdown} is
FALSE -- pills only render from HTML.}
}
\value{
Event id, or NULL on dry-run.
}
\description{
Send plain text to a Matrix room
}
\examples{
client <- list(room_id = "!default:example.org")
mx_send_text(client, "release is out", dry_run = TRUE)
\dontrun{
# A real send needs a live homeserver session:
client <- mx_client_load("myapp")
mx_send_text(client, "release is out", markdown = TRUE,
             mentions = "@jorge:example.org")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_handle_to_device}
\alias{mx_crypto_handle_to_device}
\title{Decrypt an inbound Olm to-device payload}
\usage{
mx_crypto_handle_to_device(account, my_curve25519, content, self_id = NULL,
                           self_ed25519 = NULL, devices = NULL)
}
\arguments{
\item{account}{An mx.crypto account handle.}

\item{my_curve25519}{Character. This device's Curve25519 key.}

\item{content}{The to-device \code{m.room.encrypted} content.}

\item{self_id}{Character or NULL. This user's Matrix id. When NULL the
recipient user-id check is skipped; pass it whenever it is known.}

\item{self_ed25519}{Character or NULL. This device's Ed25519 key.
Defaults to the account's own key.}

\item{devices}{List of verified devices from
\code{mx_crypto_known_devices()}, or NULL. Used to tie the payload's
claimed sender to a device whose keys were verified. The result
carries \code{sender_bound}, which is FALSE when no list is supplied
or nothing matches; an unbound sender identity is a claim, not a
fact, because anyone can open an Olm session to this device.}
}
\value{
The decrypted event (a parsed list) with a \code{sender_bound}
  flag, or NULL if it was not for us or failed the recipient checks.
}
\description{
Accepts an \code{m.room.encrypted} to-device content addressed to this
device and returns the decrypted event. When it is an \code{m.room_key},
the caller builds an inbound Megolm session from
\code{content$session_key} with \code{mx_crypto_inbound_session()}.
}
\details{
The decrypted payload is checked against this device's identity before
it is returned: a message that does not name us as recipient is
dropped, because decrypting successfully only proves it was encrypted
to our key, not that it was meant for us here.

}
\examples{
\dontrun{
ev <- mx_crypto_handle_to_device(acct, my_curve, td_content,
                                 self_id = "@me:example.org",
                                 devices = mx_crypto_known_devices(cl, uid))
if (identical(ev$type, "m.room_key") && isTRUE(ev$sender_bound)) {
  inb <- mx_crypto_inbound_session(ev$content$session_key)
}
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_room_encrypted}
\alias{mx_room_encrypted}
\title{Is a room end-to-end encrypted?}
\usage{
mx_room_encrypted(client, room = NULL, room_cache = NULL)
}
\arguments{
\item{client}{Matrix client config.}

\item{room}{Character room id/name or NULL for the default room.}

\item{room_cache}{Optional room name-to-id cache.}
}
\value{
TRUE when the room advertises an encryption algorithm,
  FALSE otherwise.
}
\description{
Resolves the room (by name, id, or the config default) and reads its
\code{m.room.encryption} state. Needs mx.api >= 0.3.0.
}
\examples{
\dontrun{
if (mx_room_encrypted(client, "secret plans")) {
    # use mx_send_encrypted() instead of mx_send_text()
}
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_encrypt_event}
\alias{mx_crypto_encrypt_event}
\title{Encrypt event content for a room (Megolm)}
\usage{
mx_crypto_encrypt_event(megolm_out, content, room_id, sender_curve25519,
                        device_id)
}
\arguments{
\item{megolm_out}{An outbound Megolm session.}

\item{content}{Named list. The plaintext event content (e.g. an
\code{m.room.message}).}

\item{room_id}{Character. Room id.}

\item{sender_curve25519}{Character. This device's Curve25519 key.}

\item{device_id}{Character. This device's id.}
}
\value{
A named list: \code{m.room.encrypted} content.
}
\description{
Returns the \code{m.room.encrypted} content to send as the event body.
}
\examples{
\dontrun{
enc <- mx_crypto_encrypt_event(megolm_out,
  list(msgtype = "m.text", body = "hi"), "!room:ex", my_curve, "DEV")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_client_config_path}
\alias{mx_client_config_path}
\title{Path to a Matrix client config file}
\usage{
mx_client_config_path(app = "mx.client", env_var = NULL)
}
\arguments{
\item{app}{Character. Application namespace for \code{tools::R_user_dir()}.}

\item{env_var}{Character or NULL. Override environment variable name.}
}
\value{
Character path.
}
\description{
Resolves the config path for an application that uses mx.client. The
default environment variable is derived from \code{app}; for example,
\code{app = "corteza"} honors \code{CORTEZA_MATRIX_CONFIG}.
}
\examples{
mx_client_config_path("myapp")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_publish_keys}
\alias{mx_crypto_publish_keys}
\title{Publish this device's identity and one-time keys}
\usage{
mx_crypto_publish_keys(client, account, store_dir, n_otks = 50L)
}
\arguments{
\item{client}{Matrix client config (needs \code{user_id},
\code{device_id}).}

\item{account}{An mx.crypto account handle.}

\item{store_dir}{Character. Crypto store directory.}

\item{n_otks}{Integer. Number of one-time keys to publish.}
}
\value{
The \code{/keys/upload} response, invisibly.
}
\description{
Builds and signs the device keys and a batch of one-time keys, uploads
them with \code{mx.api::mx_keys_upload()}, marks them published, and
persists the account. Call once after login and again to replenish
one-time keys.
}
\examples{
\dontrun{
acct <- mx_crypto_account(mx_crypto_store_dir("corteza"))
mx_crypto_publish_keys(mx_client_load(app = "corteza"), acct,
                       mx_crypto_store_dir("corteza"))
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_send_encrypted}
\alias{mx_send_encrypted}
\title{Send an end-to-end encrypted message to a room}
\usage{
mx_send_encrypted(client, account, sessions, room_id, content, store_dir,
                  recipients = NULL, member_ids = NULL)
}
\arguments{
\item{client}{Matrix client config.}

\item{account}{An mx.crypto account handle.}

\item{sessions}{A session set (see \code{mx_crypto_sessions_new()}).}

\item{room_id}{Character room id.}

\item{content}{Named list. Plaintext event content.}

\item{store_dir}{Character. Crypto store directory.}

\item{recipients}{List of recipient devices, or NULL to discover them
from \code{member_ids}.}

\item{member_ids}{Character vector of room member user ids (used when
\code{recipients} is NULL).}
}
\value{
List with \code{event_id} and the updated \code{sessions}.
}
\description{
Discovers the room members' devices (unless \code{recipients} is given),
claims one-time keys for any without an Olm session, shares the room key
over to-device, encrypts the content with Megolm, sends the
\code{m.room.encrypted} event, and persists the updated sessions.
}
\examples{
\dontrun{
res <- mx_send_encrypted(client, acct, sessions, "!r:ex",
  list(msgtype = "m.text", body = "secret"), store,
  member_ids = "@bob:example.org")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_sessions_load}
\alias{mx_crypto_sessions_load}
\title{Load a session set from the crypto store}
\usage{
mx_crypto_sessions_load(store_dir)
}
\arguments{
\item{store_dir}{Character. Crypto store directory.}
}
\value{
A session set (empty if nothing is stored yet).
}
\description{
Load a session set from the crypto store
}
\examples{
\donttest{
if (requireNamespace("mx.crypto", quietly = TRUE)) {
  s <- mx_crypto_sessions_load(file.path(tempfile(), "crypto"))
}
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_extract_invites}
\alias{mx_extract_invites}
\title{Extract pending invite room ids from a sync response}
\usage{
mx_extract_invites(sync_resp)
}
\arguments{
\item{sync_resp}{Parsed \code{/sync} response.}
}
\value{
Character vector of invited room ids.
}
\description{
Extract pending invite room ids from a sync response
}
\examples{
sync_resp <- list(rooms = list(invite = list("!inv:example.org" = list())))
mx_extract_invites(sync_resp)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_room_key_payload}
\alias{mx_crypto_room_key_payload}
\title{Encrypt a Megolm room key to one device as a to-device payload}
\usage{
mx_crypto_room_key_payload(olm_session, sender_curve25519,
                           recipient_curve25519, room_id, megolm_out,
                           sender_user_id, sender_ed25519, recipient_user_id,
                           recipient_ed25519)
}
\arguments{
\item{olm_session}{An outbound Olm session
(\code{mx.crypto::mxc_olm_create_outbound()}).}

\item{sender_curve25519}{Character. This device's Curve25519 key.}

\item{recipient_curve25519}{Character. Target device's Curve25519 key.}

\item{room_id}{Character. Room the key is for.}

\item{megolm_out}{An outbound Megolm session.}

\item{sender_user_id}{Character. This user's Matrix id.}

\item{sender_ed25519}{Character. This device's Ed25519 key.}

\item{recipient_user_id}{Character. Target user's Matrix id.}

\item{recipient_ed25519}{Character. Target device's Ed25519 key.}
}
\value{
A named list: the to-device \code{m.room.encrypted} content.
}
\description{
Wraps the outbound Megolm session's key in an \code{m.room_key} event,
Olm-encrypts it to the recipient device, and returns the
\code{m.room.encrypted} to-device content to hand to
\code{mx.api::mx_send_to_device()}.
}
\examples{
\dontrun{
content <- mx_crypto_room_key_payload(olm, my_curve, their_curve,
                                      "!room:ex", megolm_out,
                                      "@me:ex", my_ed, "@them:ex", their_ed)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_pill_mentions}
\alias{mx_pill_mentions}
\title{Turn textual @mentions into Matrix pills in formatted HTML}
\usage{
mx_pill_mentions(html, user_ids)
}
\arguments{
\item{html}{Character HTML (e.g. from \code{\link{mx_markdown_to_html}}).}

\item{user_ids}{Character Matrix user ids, such as
\code{"@jorge:example.org"}.}
}
\value{
HTML with textual \code{@localpart} occurrences replaced by
  matrix.to links. Unmatched user ids leave the HTML unchanged; they
  can still be placed in \code{m.mentions} by \code{mx_send_text()}.
}
\description{
Turn textual @mentions into Matrix pills in formatted HTML
}
\examples{
mx_pill_mentions("<p>ping @jorge</p>", "@jorge:example.org")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_resolve_room}
\alias{mx_resolve_room}
\title{Resolve a room id, name, or default room}
\usage{
mx_resolve_room(client, room = NULL, room_cache = NULL, fallback = TRUE,
                details = FALSE, quiet = FALSE)
}
\arguments{
\item{client}{Matrix client config.}

\item{room}{Character or NULL.}

\item{room_cache}{Named list or character vector mapping names to ids.}

\item{fallback}{Logical. Use \code{client$room_id} when lookup misses.}

\item{details}{Logical. Return source metadata instead of just the id.}

\item{quiet}{Logical. Suppress fallback message.}
}
\value{
Character room id, or a list when \code{details = TRUE}.
}
\description{
Resolution order:
literal room IDs beginning with \code{!}, a supplied \code{room_cache}
name-to-id map, joined-room display-name lookup, then the config's
\code{room_id} fallback.
}
\examples{
client <- list(room_id = "!default:example.org")
# Literal ids and cache hits resolve without a server round-trip:
mx_resolve_room(client, "!abc:example.org")
mx_resolve_room(client, "general",
                room_cache = list(general = "!gen:example.org"))
mx_resolve_room(client) # NULL room falls back to the config default
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_sessions_save}
\alias{mx_crypto_sessions_save}
\title{Persist a session set to the crypto store}
\usage{
mx_crypto_sessions_save(sessions, store_dir)
}
\arguments{
\item{sessions}{A session set.}

\item{store_dir}{Character. Crypto store directory.}
}
\value{
The path written, invisibly.
}
\description{
Pickles every live session (encrypted at rest with the store key) into
\code{sessions.json}. Reload with \code{mx_crypto_sessions_load()}.
}
\examples{
\donttest{
if (requireNamespace("mx.crypto", quietly = TRUE)) {
  dir <- file.path(tempfile(), "crypto")
  mx_crypto_sessions_save(mx_crypto_sessions_new(), dir)
}
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{print.mx_client_config}
\alias{print.mx_client_config}
\title{Print a Matrix client config}
\usage{
\method{print}{mx_client_config}(x, ...)
}
\arguments{
\item{x}{An \code{mx_client_config}, as returned by
\code{mx_client_load()} or \code{mx_client_from_config()}.}

\item{...}{Ignored.}
}
\value{
\code{x}, invisibly.
}
\description{
Prints the config field by field with credentials masked, so that an
interactive session, a screenshot, or a pasted bug report does not
leak the access token or password. Secret fields show whether they
are set (\code{<hidden>}) or empty (\code{<unset>}) without showing
the value. Use \code{unclass(x)} or \code{str(unclass(x))} when the
raw credentials are genuinely needed.
}
\examples{
cfg <- mx_client_from_config(list(server = "https://matrix.example.org",
                                  token = "syt_secret_value",
                                  user_id = "@bot:example.org",
                                  device_id = "DEVICEID"))
cfg
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_table_html}
\alias{mx_table_html}
\title{Render tabular data as Matrix custom HTML}
\usage{
mx_table_html(x, header = TRUE)
}
\arguments{
\item{x}{A data frame, matrix, or list coercible to a data frame.}

\item{header}{Logical. Include a header row using column names.}
}
\value{
Character HTML suitable for Matrix \code{formatted_body}.
}
\description{
Produces the conservative table shape rendered by Matrix clients such as
FluffyChat 2.6.0+: a bare \code{<table>} containing \code{<tr>},
\code{<th>}, and \code{<td>} nodes. No CSS, colspan, rowspan, or custom
attributes are emitted.
}
\examples{
mx_table_html(data.frame(A = 1:2, B = c("x", "y")))
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_inbound_session}
\alias{mx_crypto_inbound_session}
\title{Build an inbound Megolm session from a shared room key}
\usage{
mx_crypto_inbound_session(session_key)
}
\arguments{
\item{session_key}{Character. The \code{session_key} from an
\code{m.room_key} event.}
}
\value{
An inbound Megolm session.
}
\description{
Build an inbound Megolm session from a shared room key
}
\examples{
\dontrun{
inb <- mx_crypto_inbound_session(ev$content$session_key)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_client_load}
\alias{mx_client_load}
\title{Load a Matrix client config}
\usage{
mx_client_load(app = "mx.client", path = NULL,
               legacy_path = mx_client_legacy_config_path(app), env_var = NULL)
}
\arguments{
\item{app}{Character. Application namespace.}

\item{path}{Character or NULL. Explicit config path.}

\item{legacy_path}{Character or NULL. Backward-compatible fallback path.}

\item{env_var}{Character or NULL. Override environment variable name.}
}
\value{
An \code{"mx_client_config"} object.
}
\description{
Reads a JSON config. If \code{path} or the derived environment variable
is explicit, that path is authoritative. Otherwise \code{legacy_path}
is used as a compatibility fallback when present.
}
\examples{
path <- file.path(tempdir(), "matrix.json")
cfg <- mx_client_from_config(list(server = "https://matrix.example.org",
                                  token = "syt_example",
                                  user_id = "@bot:example.org",
                                  device_id = "DEVICEID"))
mx_client_save(cfg, path = path)
mx_client_load(path = path)$user_id
unlink(path)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_process_sync}
\alias{mx_crypto_process_sync}
\title{Process a sync response: store room keys, decrypt room events}
\usage{
mx_crypto_process_sync(account, sessions, sync_resp, self_curve25519,
                       self_id = NULL, devices = NULL)
}
\arguments{
\item{account}{An mx.crypto account handle.}

\item{sessions}{A session set.}

\item{sync_resp}{Parsed \code{/sync} response.}

\item{self_curve25519}{Character. This device's Curve25519 key.}

\item{self_id}{Character or NULL. This user's Matrix id, for
\code{is_self} tagging.}

\item{devices}{List of verified devices from
\code{mx_crypto_known_devices()}, or NULL. Room keys arrive over Olm
carrying a claimed sender, and anyone who can reach this device can
send one, so the claim is only worth something once it is matched
against a device whose \code{device_keys} verified. Without this
list decrypted events always report \code{sender_verified = FALSE}:
they still decrypt, but nothing attests to who sent them.}
}
\value{
List with \code{events} (decrypted, normalized) and the updated
  \code{sessions}.
}
\description{
Handles inbound to-device \code{m.room.encrypted} (Olm) messages,
storing any \code{m.room_key} as an inbound Megolm session, then
decrypts \code{m.room.encrypted} timeline events whose session is
known. Returns normalized text events in the same shape as
\code{mx_extract_text_events()}, plus the updated session set.
}
\examples{
\donttest{
if (requireNamespace("mx.crypto", quietly = TRUE)) {
  acct <- mx.crypto::mxc_account_new()
  res <- mx_crypto_process_sync(acct, mx_crypto_sessions_new(),
    list(to_device = list(events = list()), rooms = list(join = list())),
    mx.crypto::mxc_account_identity_keys(acct)$curve25519)
  length(res$events)
}
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_crypto_account}
\alias{mx_crypto_account}
\title{Load or create this client's Olm account}
\usage{
mx_crypto_account(store_dir)
}
\arguments{
\item{store_dir}{Character. Crypto store directory.}
}
\value{
An mx.crypto account handle.
}
\description{
Unpickles \code{account.pickle} from the store, or mints a fresh
account and persists it. The account holds the device's long-lived
Curve25519/Ed25519 identity keys.
}
\examples{
\donttest{
if (requireNamespace("mx.crypto", quietly = TRUE)) {
  store <- mx_crypto_store_dir("myapp", path = tempfile())
  acct <- mx_crypto_account(store)
  unlink(store, recursive = TRUE)
}
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_extract_text_events}
\alias{mx_extract_text_events}
\title{Extract text message events from a sync response}
\usage{
mx_extract_text_events(sync_resp, self_id, msgtypes = "m.text")
}
\arguments{
\item{sync_resp}{Parsed \code{/sync} response.}

\item{self_id}{Current user's Matrix id.}

\item{msgtypes}{Character vector of message types to include.}
}
\value{
List of normalized event records, each carrying
  \code{room_id}, \code{event_id}, \code{sender}, \code{is_self},
  \code{body}, \code{msgtype}, \code{ts} (the event's
  \code{origin_server_ts}, in milliseconds since the epoch, or NULL
  when the server omits it), and \code{mentions}.
}
\description{
Walks joined-room timeline events and returns normalized text-message
records. Self events are retained and tagged with \code{is_self}.
}
\examples{
sync_resp <- list(rooms = list(join = list("!room:example.org" = list(
    timeline = list(events = list(list(type = "m.room.message",
        event_id = "$1", sender = "@alice:example.org",
        content = list(msgtype = "m.text", body = "hello"))))))))
mx_extract_text_events(sync_resp, self_id = "@bot:example.org")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_client_configure}
\alias{mx_client_configure}
\title{Configure and save a Matrix client}
\usage{
mx_client_configure(server, user, password, room, app = "mx.client",
                    path = NULL, device_id = NULL, extra = list())
}
\arguments{
\item{server}{Character. Homeserver base URL.}

\item{user}{Character. Matrix user localpart or full MXID.}

\item{password}{Character. Account password.}

\item{room}{Character. Room ID or alias to join.}

\item{app}{Character. Application namespace.}

\item{path}{Character or NULL. Explicit destination path.}

\item{device_id}{Character or NULL. Existing device id to reuse.}

\item{extra}{Named list. Additional fields to save.}
}
\value{
Saved \code{"mx_client_config"}, invisibly.
}
\description{
Logs in with \pkg{mx.api}, joins or records the target room, and writes
a reusable local config. Extra fields are merged into the saved config
so applications can persist their own defaults without reimplementing
login.
}
\examples{
\dontrun{
# Needs a live homeserver and account credentials.
mx_client_configure("https://matrix.example.org", "bot", "secret",
                    room = "#general:example.org", app = "myapp")
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_client_legacy_config_path}
\alias{mx_client_legacy_config_path}
\title{Legacy Matrix config path for an application}
\usage{
mx_client_legacy_config_path(app = "mx.client")
}
\arguments{
\item{app}{Character. Application namespace.}
}
\value{
Character path or NULL.
}
\description{
Currently only \code{app = "corteza"} has a historical path:
\code{~/.corteza/matrix.json}.
}
\examples{
mx_client_legacy_config_path("corteza")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_with_relogin}
\alias{mx_with_relogin}
\title{Run a client operation, re-logging in once on an expired token}
\usage{
mx_with_relogin(client, fn, save = TRUE)
}
\arguments{
\item{client}{Matrix client config with \code{password}.}

\item{fn}{Function taking a client config.}

\item{save}{Logical. Persist the refreshed config on relogin.}
}
\value{
\code{fn}'s return value.
}
\description{
Calls \code{fn(client)}; if it fails with the server's invalid-token
error (\code{M_UNKNOWN_TOKEN}, signalled as a classed condition by
mx.api >= 0.3.0), re-logs in via \code{\link{mx_client_relogin}} and
retries once with the refreshed client. Any other error propagates.
}
\examples{
\dontrun{
mx_with_relogin(client, function(cl) {
    mx_send_text(cl, "still here after a token rotation")
})
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{mx_send_media}
\alias{mx_send_media}
\title{Send a media file to a Matrix room}
\usage{
mx_send_media(client, path, room = NULL, body = basename(path), msgtype = NULL,
              content_type = NULL, info = list(), room_cache = NULL,
              dry_run = FALSE)
}
\arguments{
\item{client}{Matrix client config.}

\item{path}{Character. Path to the file to upload.}

\item{room}{Character room id/name or NULL for the default room.}

\item{body}{Character. Message body / filename shown by clients.}

\item{msgtype}{Character or NULL. NULL derives it from the MIME type.}

\item{content_type}{Character or NULL. MIME type override for files
whose extension guesses wrong (tempfiles, odd extensions); NULL
guesses from the extension.}

\item{info}{List. Extra fields merged into the media \code{info}.}

\item{room_cache}{Optional room name-to-id cache.}

\item{dry_run}{Logical. Print instead of uploading/sending.}
}
\value{
Event id, or NULL on dry-run.
}
\description{
Client-layer wrapper over \code{mx.api::mx_send_media()}: resolves the
room by name (or falls back to the config's default room), builds the
session from the client config, and uploads + posts in one call. The
msgtype is derived from the file's MIME type unless given.
}
\details{
If you attach mx.api and mx.client together, namespace-qualify -- the
two packages export an \code{mx_send_media} each (session-first there,
client-first here).

}
\examples{
client <- list(room_id = "!default:example.org")
png <- file.path(tempdir(), "plot.png")
file.create(png)
mx_send_media(client, png, dry_run = TRUE)
unlink(png)
}
